import argparse
import os
import datetime
import re
import shutil
import pandas as pd
import subprocess
import rpy2.robjects as robjects
from rpy2.robjects.packages import importr
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_100_summarize_outputs")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument('--force', 
    action='store_true', 
    help='Force execution of all steps, even if final_annotation_output_file exists.',
    )

class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()

def find_max_module(args):
    max_module = 0
    max_module_folder = ""
    target_directories = ['01_GENOMAD', '02_CHECK_V', '03_CLUSTERING', '04_READ_MAPPING', '05_VOTU_TABLES', '06_FUNCTIONAL_ANNOTATION', '07_BINNING']

    for root, dirs, files in os.walk(args['input']):
        for d in dirs:
            if d in target_directories:
                module_number = int(d.split("_")[0])
                if module_number > max_module and module_number < 99:
                    max_module = module_number
                    max_module_folder = d

    return max_module, max_module_folder

def find_summary_reports(args, max_module_folder):
    target_directories = [os.path.join(args['input']), '01_GENOMAD', '02_CHECK_V', '03_CLUSTERING', '04_READ_MAPPING', '05_VOTU_TABLES', '06_FUNCTIONAL_ANNOTATION', '07_BINNING']
    summary_reports_dict = {}
    
    # Include the base directory in the search for summary reports
    all_directories = [args['input']] + [os.path.join(args['input'], directory) for directory in target_directories]
    
    for directory_path in all_directories:
        summary_reports = []
        if os.path.exists(directory_path):
            for root, dirs, files in os.walk(directory_path):
                for file in files:
                    if file.endswith("_Summary_Report.txt"):
                        summary_reports.append(os.path.join(root, file))
                if os.path.basename(directory_path) == max_module_folder:  # Check if it's the max_module_folder
                    break  # Stop searching further
            summary_reports_dict[os.path.basename(directory_path)] = summary_reports

    num_summary_reports = len(summary_reports_dict['01_GENOMAD'])

    return summary_reports_dict, num_summary_reports

def create_output_folder(args):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the output directory with today's date to store the summary files...\033[0m")

    start_time = datetime.datetime.now()
    formatted_start_time = start_time.strftime("%Y-%m-%d-%Hh%Mmin%Ssec")

    output_folder = os.path.join(args['input'], "100_SUMMARIZED_OUTPUTS", formatted_start_time)
    
    # Check if --force argument is provided
    if args['force']:
        # Delete existing output_folder if it exists
        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)
    
    # Create the output directory
    os.makedirs(output_folder, exist_ok=True)
    output_filename = os.path.join(output_folder, f"{formatted_start_time}_MVP_100_Summary_Report.txt")
    
    return output_folder, output_filename

def extract_module_info(report_list):
    module_info_list = []
    for report in report_list:
        with open(report, 'r') as file:
            module_info = []
            lines_after_running_time = False
            for line in file:
                if line.startswith('*') or '.py' in line:
                    module_info.append(line.strip())
                # Check if the report is in the specified list
                if os.path.basename(report) in ['MVP_03_Summary_Report.txt', 'MVP_05_Summary_Report.txt', 'MVP_06_Summary_Report.txt', 'MVP_07_Summary_Report.txt']:
                    if lines_after_running_time:
                        module_info.append(line.strip())
                    elif line.startswith('Running Time:'):
                        lines_after_running_time = True
        module_info_list.append("\n".join(module_info))
    return module_info_list

def extract_running_times(report_list, trigger_line):
    running_times = []
    for report in report_list:
        with open(report, 'r') as file:
            found_trigger_line = False
            for line in file:
                if found_trigger_line and line.startswith('Running Time'):
                    match = re.search(r'(\d+.\d+) seconds', line)
                    if match:
                        running_time_str = match.group(1)
                        running_time = float(running_time_str)
                        running_times.append(running_time)
                elif line.startswith(trigger_line):
                    found_trigger_line = True
    return running_times

def calculate_MVP_total_running_time(summary_reports_dict):
    total_running_time = 0
    for directory, data in summary_reports_dict.items():
        total_running_time += sum(extract_running_times(data, directory.split('_')[0]))
    return total_running_time

def format_time(seconds):
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return "{:02}:{:02}:{:02}".format(int(hours), int(minutes), int(seconds))

def generate_summary_report(output_filename, max_module, max_module_folder, num_summary_reports, summary_reports_dict):
    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the final summary report...\033[0m")

    with open(output_filename, "w") as f:
        f.write("****************************************************************************\n")
        f.write("************            MVP Module 100 Summary Report            ***********\n")
        f.write("****************************************************************************\n")
        f.write(f"Number of samples processed by MVP: {num_summary_reports}\n")
        f.write(f"Number of MVP modules ran: 0{max_module} ({max_module_folder})\n")
        total_running_time = calculate_MVP_total_running_time(summary_reports_dict)
        f.write(f"Total MVP running time (hours:minutes:seconds): {format_time(total_running_time)}")
        
        # Print running times and means
        for directory, data in summary_reports_dict.items():
            total_seconds = sum(extract_running_times(data, directory.split('_')[0]))
            total_reports = len(data)
            if total_reports > 0:
                module_info = extract_module_info(data)
                f.write("\n")
                f.write(f"\n{module_info[0]}")  # Write only the first 4 lines
                f.write("\n")
                f.write(f"\nTotal running time (hours:minutes:seconds): {format_time(total_seconds)}")
                if total_reports > 1:
                    mean_seconds = total_seconds / total_reports
                    mean_hours = mean_seconds // 3600
                    mean_minutes = (mean_seconds % 3600) // 60
                    mean_seconds %= 60
                    f.write(f"\nMean running time per summary report: {format_time(mean_hours * 3600 + mean_minutes * 60 + mean_seconds)}")

def copy_additional_files(args, max_module_folder, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Copying main output files in {output_folder}...\033[0m")

    source_files = [
        "MVP_03_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Representative_Virus_Proviruses_Quality_Summary.tsv",
        "MVP_05_All_Sample_Filtered_conservative_Representative_Virus_Proviruses_vOTU_RPKM_Table.tsv",
        "_Filtered.tsv",
        "MVP_07_Merged_vRhyme_Outputs_Filtered_conservative_best_vBins_Representative_Unbinned_vOTUs_geNomad_CheckV_Summary_read_mapping_information_RPKM_Table.tsv"
    ]

    if max_module_folder == '03_CLUSTERING':
        source_files_to_copy = [source_files[0]]
    elif max_module_folder == '05_VOTU_TABLES':
        source_files_to_copy = source_files[:2]
    elif max_module_folder == '06_FUNCTIONAL_ANNOTATION':
        source_files_to_copy = source_files[:3]
    elif max_module_folder == '07_BINNING':
        source_files_to_copy = source_files

    # Get all files containing "_Filtered.tsv" in their names
    additional_files = []
    for root, _, files in os.walk(args['input']):
        for file in files:
            if "_Filtered.tsv" in file:
                additional_files.append(file)

    # Combine the initial set of files and additional files
    source_files += additional_files

    for file in source_files:
        destination_path = os.path.join(output_folder, file)
        # Only copy the file if it doesn't already exist in the output folder
        if not os.path.exists(destination_path):
            # Search for the file in base_directory and its subfolders
            for root, _, files in os.walk(args['input']):
                if file in files:
                    source_path = os.path.join(root, file)
                    shutil.copy(source_path, output_folder)
                    # Convert to CSV if the file is "_Filtered.tsv"
                    if "_Filtered.tsv" in file:
                      with open(destination_path, 'r') as f:
                          data = f.read()
                      with open(destination_path, 'w') as f:
                          f.write(data.replace(',', ';'))                        

                      csv_file = os.path.splitext(destination_path)[0] + ".csv"
                      os.rename(destination_path, csv_file)
                      # Convert to CSV format with comma as separator
                      with open(csv_file, 'r') as f:
                          data = f.read()
                      with open(csv_file, 'w') as f:
                          f.write(data.replace('\t', ','))  # Replace tabs with commas
                    break  # Stop searching if the file is found

def install_r_packages():
    print(f"\n\033[1m{step_counter.print_main_step()}: Installing R packages if not already installed...\033[0m")

    packages_to_install = ["ggplot2", "dplyr", "cowplot", "readr", "complexupset", "stringr", "reshape2", "vegan"]

    installed_packages = subprocess.run('conda list', shell=True, capture_output=True, text=True).stdout

    for package in packages_to_install:
        if f'r-{package}' not in installed_packages:
            print(f"Installing {package}...")
            subprocess.run(f"conda install -c r r-{package} -y", shell=True)

def find_R_tables(output_folder):
    files = os.listdir(output_folder)
    votu_table_file_name = None
    for file in files:
        if file.startswith('MVP_07_'):
            votu_table_file_name = file
            break  # Stop iteration if MVP_07 is found

    if not votu_table_file_name:  # If MVP_07_ file not found
        for file in files:
            if file.startswith('MVP_05_'):
                votu_table_file_name = file
                break
            elif file.startswith('MVP_03_'):
                votu_table_file_name = file
            
    functional_annotation_table_file_path = None
    for file in files:
        if file.startswith('MVP_06_All_Sample_Filtered_Conservative_'):
            functional_annotation_file_name = os.path.join(file)
            break
        elif file.startswith('MVP_06_All_Sample_Filtered_Relaxed_'):
            functional_annotation_file_name = os.path.join(file)

    votu_table_file_path = os.path.join(output_folder, votu_table_file_name)
    functional_annotation_table_file_path = os.path.join(output_folder, functional_annotation_file_name)

    return votu_table_file_name, votu_table_file_path, functional_annotation_table_file_path

def create_module_03_figures(votu_table_file_path, metadata_file_path, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating figures for Module 03 using R...\033[0m")

    r_code = """
library(ggplot2)
library(dplyr)
library(cowplot)
library(readr)
library(ComplexUpset)
library(stringr)
library(reshape2)
library(vegan)

data = read.table("{votu_table_file_path}", h=T, sep="\t", dec=".")

checkv_data <- data %>%
  count(checkv_quality = checkv_quality) %>%
  mutate(checkv_quality = factor(checkv_quality, levels = c("Complete", "High-quality", "Medium-quality", "Low-quality", "Not-determined"))) %>%
  mutate(Factor = "checkv_quality")

# Create the box plot
p_conservative_filtered_vOTUs_CheckV <- ggplot(checkv_data, aes(x = Factor, y = n, fill = checkv_quality)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.8) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3", "#9D9D9D", "#686868", "#333333"), 
                    limits = c("Not-determined", "Low-quality", "Medium-quality", "High-quality", "Complete"),
                    name="CheckV quality") +
  coord_flip() +
  labs(y = "Number", x = "") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

length_data <- data %>%
  mutate(Length_Category = case_when(
    virus_length < 10000 ~ "<10 kb",
    virus_length >= 10000 & virus_length <= 50000 ~ "10-50 kb",
    TRUE ~ ">50 kb")) %>%
  count(Length_Category) %>%
  mutate(Length_Category = factor(Length_Category, levels = c("<10 kb", "10-50 kb", ">50 kb"))) %>%
  mutate(Factor = "Length")

p_conservative_filtered_vOTUs_length <- ggplot(length_data, aes(x = Length_Category, y = n)) +
  geom_bar(stat="identity", position= position_dodge(preserve = "single"), fill = '#D3D3D3', color = "black", size = 1) + 
  theme_minimal() +
  coord_flip() +
  labs(y = "Number", x = "Genome length") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

data <- data %>%
  mutate(Taxonomy = sapply(strsplit(as.character(taxonomy), ";"), function(x) {{
    level_5 <- ifelse(length(x) >= 5, x[5], tail(x, 1))
    if (level_5 == "Viruses") {{
      return("Unclassified")
    }} else {{
      return(level_5)
    }}
  }}))

taxa_count_data <- data %>%
  count(Taxonomy = data$Taxonomy) %>%
  mutate(Factor = "Taxonomy")
head(taxa_count_data)

# Define the color palette based on the presence of "Unclassified"
if ("Unclassified" %in% taxa_count_data$Taxonomy) {{
  # If "Unclassified" exists, use one less color than the number of unique taxa
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)) - 1)
  # Assign #D3D3D3 to Unclassified
  color_palette <- c(color_palette, "#D3D3D3")
}} else {{
  # If "Unclassified" doesn't exist, use the number of unique taxa as the number of colors
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)))
}}  

# Create the ggplot
p_taxa_vOTUs_counts <- ggplot(taxa_count_data, aes(x = Factor, y = n, fill = Taxonomy)) +
  geom_bar(position = "fill", stat = "identity", color = "black", size = 1) + 
  scale_fill_manual(values = color_palette) +
  labs(x = "", y = "Percentage (%)", fill = "Taxonomy") + 
  theme_minimal() +
  coord_flip() +
  scale_y_continuous(labels = scales::percent_format()) +
  theme(legend.justification = c(0, 0.5), 
        axis.text = element_text(color = "black", size = 14),
        axis.title = element_text(color = "black", size = 14),
        axis.ticks = element_line(size = 0.8, color = "black"),
        axis.ticks.length = unit(0.2, "cm"))

spacer <-  ggplot() +
                geom_blank(fill = "white") +  # Set the background color to white
                theme_void()

title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 03 (Clustering)", fontface='bold')

combined_plots <- plot_grid(title, p_conservative_filtered_vOTUs_CheckV, p_conservative_filtered_vOTUs_length, p_taxa_vOTUs_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.4, 0.6, 0.6), labels=c('','A','B', 'C'))

ggsave("{output_folder}/Summarize_output_plots.pdf", plot = combined_plots, width = 8, height = 8)
    """.format(votu_table_file_path=votu_table_file_path, metadata_file_path=metadata_file_path, output_folder=output_folder)

    # Run the R code
    robjects.r(r_code)

def create_module_03_06_figures(votu_table_file_path, metadata_file_path, functional_annotation_table_file_path, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating figures for Module 03 and Module 06 using R...\033[0m")


    r_code = """
library(ggplot2)
library(dplyr)
library(cowplot)
library(readr)
library(ComplexUpset)
library(stringr)
library(reshape2)
library(vegan)

data = read.table("{votu_table_file_path}", h=T, sep="\t", dec=".")

checkv_data <- data %>%
  count(checkv_quality = checkv_quality) %>%
  mutate(checkv_quality = factor(checkv_quality, levels = c("Complete", "High-quality", "Medium-quality", "Low-quality", "Not-determined"))) %>%
  mutate(Factor = "checkv_quality")

# Create the box plot
p_conservative_filtered_vOTUs_CheckV <- ggplot(checkv_data, aes(x = Factor, y = n, fill = checkv_quality)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.8) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3", "#9D9D9D", "#686868", "#333333"), 
                    limits = c("Not-determined", "Low-quality", "Medium-quality", "High-quality", "Complete"),
                    name="CheckV quality") +
  coord_flip() +
  labs(y = "Number", x = "") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

length_data <- data %>%
  mutate(Length_Category = case_when(
    virus_length < 10000 ~ "<10 kb",
    virus_length >= 10000 & virus_length <= 50000 ~ "10-50 kb",
    TRUE ~ ">50 kb")) %>%
  count(Length_Category) %>%
  mutate(Length_Category = factor(Length_Category, levels = c("<10 kb", "10-50 kb", ">50 kb"))) %>%
  mutate(Factor = "Length")

p_conservative_filtered_vOTUs_length <- ggplot(length_data, aes(x = Length_Category, y = n)) +
  geom_bar(stat="identity", position= position_dodge(preserve = "single"), fill = '#D3D3D3', color = "black", size = 1) + 
  theme_minimal() +
  coord_flip() +
  labs(y = "Number", x = "Genome length") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

data <- data %>%
  mutate(Taxonomy = sapply(strsplit(as.character(taxonomy), ";"), function(x) {{
    level_5 <- ifelse(length(x) >= 5, x[5], tail(x, 1))
    if (level_5 == "Viruses") {{
      return("Unclassified")
    }} else {{
      return(level_5)
    }}
  }}))

taxa_count_data <- data %>%
  count(Taxonomy = data$Taxonomy) %>%
  mutate(Factor = "Taxonomy")
head(taxa_count_data)

# Define the color palette based on the presence of "Unclassified"
if ("Unclassified" %in% taxa_count_data$Taxonomy) {{
  # If "Unclassified" exists, use one less color than the number of unique taxa
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)) - 1)
  # Assign #D3D3D3 to Unclassified
  color_palette <- c(color_palette, "#D3D3D3")
}} else {{
  # If "Unclassified" doesn't exist, use the number of unique taxa as the number of colors
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)))
}}  

# Create the ggplot
p_taxa_vOTUs_counts <- ggplot(taxa_count_data, aes(x = Factor, y = n, fill = Taxonomy)) +
  geom_bar(position = "fill", stat = "identity", color = "black", size = 1) + 
  scale_fill_manual(values = color_palette) +
  labs(x = "", y = "Percentage (%)", fill = "Taxonomy") + 
  theme_minimal() +
  coord_flip() +
  scale_y_continuous(labels = scales::percent_format()) +
  theme(legend.justification = c(0, 0.5), 
        axis.text = element_text(color = "black", size = 14),
        axis.title = element_text(color = "black", size = 14),
        axis.ticks = element_line(size = 0.8, color = "black"),
        axis.ticks.length = unit(0.2, "cm"))

if ("_ADS_" %in% str_extract("{functional_annotation_table_file_path}", "_ADS_")) {{
    data_M06 <- readr::read_csv("{functional_annotation_table_file_path}", col_types = cols(ADS_Defense_type = col_character()))
    data_M06$PHROGS_Category <- ifelse(is.na(data_M06$PHROGS_Category), "unknown", data_M06$PHROGS_Category)
    data_M06$PFAM_Annotation <- ifelse(is.na(data_M06$PFAM_Annotation), "unknown", data_M06$PFAM_Annotation)
    data_M06$ADS_Defense_type <- ifelse(is.na(data_M06$ADS_Defense_type), "unknown", data_M06$ADS_Defense_type)
    # Initialize lists for each annotation method
    ADS_list <- data.frame(Viral_gene_ID = character(0))
    Pfam_list <- data.frame(Viral_gene_ID = character(0))
    PHROGS_list <- data.frame(Viral_gene_ID = character(0))
    geNomad_list <- data.frame(Viral_gene_ID = character(0))

    # Filter data for each annotation method and populate the lists
    ADS_list <- data_M06 %>%
    filter(ADS_Defense_type != "unknown" & ADS_Defense_type != "Unknown") %>%
    select(Viral_gene_ID)

    # Filter out rows based on PFAM_Annotation column
    Pfam_list <- data_M06 %>%
    filter(!grepl(paste(c("unknown", "Unknown", "Uncharacterized"), collapse = "|"), PFAM_Annotation)) %>%
    select(Viral_gene_ID)

    PHROGS_list <- data_M06 %>%
    filter(PHROGS_Category != "unknown" & PHROGS_Category != "Unknown") %>%
    select(Viral_gene_ID)

    geNomad_list <- data_M06 %>%
    filter(GENOMAD_Annotation != "unknown" & GENOMAD_Annotation != "Unknown") %>%
    select(Viral_gene_ID)

    # Combine all lists into a single data frame
    upset_data <- data.frame(Viral_gene_ID = data_M06$Viral_gene_ID,
                            ADS = as.integer(data_M06$Viral_gene_ID %in% ADS_list$Viral_gene_ID),
                            Pfam = as.integer(data_M06$Viral_gene_ID %in% Pfam_list$Viral_gene_ID),
                            PHROGS = as.integer(data_M06$Viral_gene_ID %in% PHROGS_list$Viral_gene_ID),
                            geNomad = as.integer(data_M06$Viral_gene_ID %in% geNomad_list$Viral_gene_ID))

    p_annotation_upset <- upset(
                            upset_data, 
                            name='Group',
                            width_ratio=0.3,
                            c('ADS', 'Pfam', 'PHROGS', 'geNomad'),
                            queries = list(upset_query(set = 'ADS', fill = '#EFE32A'),
                                            upset_query(set = 'Pfam', fill = '#CB1800'),
                                            upset_query(set = 'PHROGS', fill = '#00A87E'),
                                            upset_query(set = 'geNomad', fill = '#61BCEC')),
                            base_annotations=list('Shared genes between groups'=(intersection_size(counts=T, col = "black", 
                                                                                                    bar_number_threshold=1, width=0.8) +
                                                                                    scale_y_continuous(expand=expansion(mult=c(0, 0.05))) + 
                                                                                    theme(panel.grid=element_blank(),
                                                                                        panel.background = element_rect(fill = 'white'),
                                                                                        axis.text.x=element_blank(),
                                                                                        axis.text.y = element_text(color="black", size=14),
                                                                                        axis.title = element_text(color="black", size=14),
                                                                                        axis.title.x=element_blank(),
                                                                                        axis.line=element_line(colour='black')))),
                            set_sizes=(upset_set_size(geom=geom_bar(width=0.6, col = "black"),
                                                        position = 'right') + 
                                        geom_text(aes(label=after_stat(count)), hjust=-0.1, stat='count') +
                                        ylab("Nb of genes per group") +
                                        expand_limits(y=7000) +
                                        theme(panel.background =element_blank(),
                                                axis.line.x=element_line(colour='black'),
                                                axis.ticks.y=element_blank(),
                                                axis.text.y=element_blank(),
                                                axis.text.x = element_text(color="black", size=14),
                                                axis.title.x = element_text(color="black", size=14),
                                                axis.title.y=element_blank())),
                            theme = list(legend.text.align = NULL),
                            matrix=(intersection_matrix(geom=geom_point(shape='circle filled', size=3.5, stroke=0.45)) + 
                                        ylab("Database") +
                                        theme(panel.background =element_blank(),
                                            axis.line=element_blank(),
                                            axis.ticks=element_blank(),
                                            axis.title=element_text(color="black", size=14),
                                            axis.text.y = element_text(color="black", size=14),
                                            axis.text.x=element_blank())),
                            stripes=upset_stripes(geom=geom_segment(linewidth=12), colors=c('grey95', 'white')),
                            sort_sets='descending',
                            sort_intersections='descending')
}} else {{
    data_M06 <- readr::read_csv("{functional_annotation_table_file_path}")
    data_M06$PHROGS_Category <- ifelse(is.na(data_M06$PHROGS_Category), "unknown", data_M06$PHROGS_Category)
    data_M06$PFAM_Annotation <- ifelse(is.na(data_M06$PFAM_Annotation), "unknown", data_M06$PFAM_Annotation)
    # Initialize lists for each annotation method
    Pfam_list <- data.frame(Viral_gene_ID = character(0))
    PHROGS_list <- data.frame(Viral_gene_ID = character(0))
    geNomad_list <- data.frame(Viral_gene_ID = character(0))

    # Filter out rows based on PFAM_Annotation column
    Pfam_list <- data_M06 %>%
    filter(!grepl(paste(c("unknown", "Unknown", "Uncharacterized"), collapse = "|"), PFAM_Annotation)) %>%
    select(Viral_gene_ID)

    PHROGS_list <- data_M06 %>%
    filter(PHROGS_Category != "unknown" & PHROGS_Category != "Unknown") %>%
    select(Viral_gene_ID)

    geNomad_list <- data_M06 %>%
    filter(GENOMAD_Annotation != "unknown" & GENOMAD_Annotation != "Unknown") %>%
    select(Viral_gene_ID)

    # Combine all lists into a single data frame
    upset_data <- data.frame(Viral_gene_ID = data_M06$Viral_gene_ID,
                            Pfam = as.integer(data_M06$Viral_gene_ID %in% Pfam_list$Viral_gene_ID),
                            PHROGS = as.integer(data_M06$Viral_gene_ID %in% PHROGS_list$Viral_gene_ID),
                            geNomad = as.integer(data_M06$Viral_gene_ID %in% geNomad_list$Viral_gene_ID))

    p_annotation_upset <- upset(
                            upset_data, 
                            name='Group',
                            width_ratio=0.3,
                            c('Pfam', 'PHROGS', 'geNomad'),
                            queries = list(upset_query(set = 'Pfam', fill = '#CB1800'),
                                            upset_query(set = 'PHROGS', fill = '#00A87E'),
                                            upset_query(set = 'geNomad', fill = '#61BCEC')),
                            base_annotations=list('Shared genes between groups'=(intersection_size(counts=T, col = "black", 
                                                                                                    bar_number_threshold=1, width=0.8) +
                                                                                    scale_y_continuous(expand=expansion(mult=c(0, 0.05))) + 
                                                                                    theme(panel.grid=element_blank(),
                                                                                        panel.background = element_rect(fill = 'white'),
                                                                                        axis.text.x=element_blank(),
                                                                                        axis.text.y = element_text(color="black", size=14),
                                                                                        axis.title = element_text(color="black", size=14),
                                                                                        axis.title.x=element_blank(),
                                                                                        axis.line=element_line(colour='black')))),
                            set_sizes=(upset_set_size(geom=geom_bar(width=0.6, col = "black"),
                                                        position = 'right') + 
                                        geom_text(aes(label=after_stat(count)), hjust=-0.1, stat='count') +
                                        ylab("Nb of genes per group") +
                                        theme(panel.background =element_blank(),
                                                axis.line.x=element_line(colour='black'),
                                                axis.ticks.y=element_blank(),
                                                axis.text.y=element_blank(),
                                                axis.text.x = element_text(color="black", size=14),
                                                axis.title.x = element_text(color="black", size=14),
                                                axis.title.y=element_blank())),
                            theme = list(legend.text.align = NULL),
                            matrix=(intersection_matrix(geom=geom_point(shape='circle filled', size=3.5, stroke=0.45)) + 
                                        ylab("Database") +
                                        theme(panel.background =element_blank(),
                                            axis.line=element_blank(),
                                            axis.ticks=element_blank(),
                                            axis.title=element_text(color="black", size=14),
                                            axis.text.y = element_text(color="black", size=14),
                                            axis.text.x=element_blank())),
                            stripes=upset_stripes(geom=geom_segment(linewidth=12), colors=c('grey95', 'white')),
                            sort_sets='descending',
                            sort_intersections='descending')
}}

spacer <-  ggplot() +
                geom_blank(fill = "white") +  # Set the background color to white
                theme_void()

title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 03 (Clustering)and 06 (Functionnal annotation)", fontface='bold')

plot_col_1 <- plot_grid(title, p_conservative_filtered_vOTUs_CheckV, p_conservative_filtered_vOTUs_length, p_taxa_vOTUs_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.4, 0.6, 0.6), labels=c('','A','B', 'C'))
plot_col_2 <- plot_grid(spacer,  p_annotation_upset, spacer,  rel_heights = c(0.2, 0.8, 0.3),
                            ncol = 1, labels=c('','D',''))
combined_plots <- plot_grid(plot_col_1, plot_col_2, align = "hv", ncol = 2)

ggsave("{output_folder}/Summarize_output_plots.pdf", plot = combined_plots, width = 20, height = 12)
    """.format(votu_table_file_path=votu_table_file_path, metadata_file_path=metadata_file_path, functional_annotation_table_file_path=functional_annotation_table_file_path, output_folder=output_folder)

    # Run the R code
    robjects.r(r_code)

def create_module_05_figures(votu_table_file_path, metadata_file_path, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating figures for Module 05 using R...\033[0m")

    r_code = """
library(ggplot2)
library(dplyr)
library(cowplot)
library(readr)
library(ComplexUpset)
library(stringr)
library(reshape2)
library(vegan)

data = read.table("{votu_table_file_path}", h=T, sep="\t", dec=".")

checkv_data <- data %>%
  count(checkv_quality = checkv_quality) %>%
  mutate(checkv_quality = factor(checkv_quality, levels = c("Complete", "High-quality", "Medium-quality", "Low-quality", "Not-determined"))) %>%
  mutate(Factor = "checkv_quality")

# Create the box plot
p_conservative_filtered_vOTUs_CheckV <- ggplot(checkv_data, aes(x = Factor, y = n, fill = checkv_quality)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.8) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3", "#9D9D9D", "#686868", "#333333"), 
                    limits = c("Not-determined", "Low-quality", "Medium-quality", "High-quality", "Complete"),
                    name="CheckV quality") +
  coord_flip() +
  labs(y = "Number", x = "") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

length_data <- data %>%
  mutate(Length_Category = case_when(
    virus_length < 10000 ~ "<10 kb",
    virus_length >= 10000 & virus_length <= 50000 ~ "10-50 kb",
    TRUE ~ ">50 kb")) %>%
  count(Length_Category) %>%
  mutate(Length_Category = factor(Length_Category, levels = c("<10 kb", "10-50 kb", ">50 kb"))) %>%
  mutate(Factor = "Length")

p_conservative_filtered_vOTUs_length <- ggplot(length_data, aes(x = Length_Category, y = n)) +
  geom_bar(stat="identity", position= position_dodge(preserve = "single"), fill = '#D3D3D3', color = "black", size = 1) + 
  theme_minimal() +
  coord_flip() +
  labs(y = "Number", x = "Genome length") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

data <- data %>%
  mutate(Taxonomy = sapply(strsplit(as.character(taxonomy), ";"), function(x) {{
    level_5 <- ifelse(length(x) >= 5, x[5], tail(x, 1))
    if (level_5 == "Viruses") {{
      return("Unclassified")
    }} else {{
      return(level_5)
    }}
  }}))

taxa_count_data <- data %>%
  count(Taxonomy = data$Taxonomy) %>%
  mutate(Factor = "Taxonomy")
head(taxa_count_data)

# Define the color palette based on the presence of "Unclassified"
if ("Unclassified" %in% taxa_count_data$Taxonomy) {{
  # If "Unclassified" exists, use one less color than the number of unique taxa
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)) - 1)
  # Assign #D3D3D3 to Unclassified
  color_palette <- c(color_palette, "#D3D3D3")
}} else {{
  # If "Unclassified" doesn't exist, use the number of unique taxa as the number of colors
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)))
}}  

# Create the ggplot
p_taxa_vOTUs_counts <- ggplot(taxa_count_data, aes(x = Factor, y = n, fill = Taxonomy)) +
  geom_bar(position = "fill", stat = "identity", color = "black", size = 1) + 
  scale_fill_manual(values = color_palette) +
  labs(x = "", y = "Percentage (%)", fill = "Taxonomy") + 
  theme_minimal() +
  coord_flip() +
  scale_y_continuous(labels = scales::percent_format()) +
  theme(legend.justification = c(0, 0.5), 
        axis.text = element_text(color = "black", size = 14),
        axis.title = element_text(color = "black", size = 14),
        axis.ticks = element_line(size = 0.8, color = "black"),
        axis.ticks.length = unit(0.2, "cm"))

        
metadata = read.table("{metadata_file_path}", h=T, sep="\t", dec=".")

if ("Variable" %in% colnames(metadata)) {{
    metadata_subset <- metadata[, c('Sample', 'Variable')]
    selected_RPKM_columns <- c("virus_id", grep('^RPKM_', names(data), value = TRUE))
    RPKM_data <- data[selected_RPKM_columns]
    colnames(RPKM_data) <- gsub('^RPKM_', '', colnames(RPKM_data))
    RPKM_data_t <- dcast(melt(RPKM_data, id.vars = "virus_id"), variable ~ virus_id)
    RPKM_data_t <- RPKM_data_t %>% rename(Sample = variable)
    RPKM_data_t_metadata <- merge(x = metadata_subset, y = RPKM_data_t, by = "Sample")
    RPKM_data_t_metadata_bio <- RPKM_data_t_metadata[,-c(1,2)]

    MDS <- cmdscale(vegdist(RPKM_data_t_metadata_bio, method = "bray"), k = 3, eig = T, add = T )
    variance_explained <- round(MDS$eig * 100 / sum(MDS$eig), 1)
    variance_explained_first_two <- head(variance_explained, 2)
    nMDS1_label <- paste0("nMDS 1 (", variance_explained_first_two[1], "%)")
    nMDS2_label <- paste0("nMDS 2 (", variance_explained_first_two[2], "%)")
    RPKM_data_t_metadata_bio_dist <- vegdist(RPKM_data_t_metadata_bio, method = "bray")
    RPKM_data_t_metadata_bio_dist_MDS <- metaMDS(RPKM_data_t_metadata_bio_dist, distance = "bray", k = 3, trymax=1000)
    nmds_coordinates <- data.frame(RPKM_data_t_metadata_bio_dist_MDS$points)
    nmds_data <- cbind(nmds_coordinates, Variable = RPKM_data_t_metadata$Variable)

    color_palette <- colorRampPalette(c("#A36B2B", "#C6AA74", "#EDEAC2", "#94B9A7", "#2686A0"))(length(unique(RPKM_data_t_metadata$Variable)))

    p_conservative_filtered_vOTUs_nMDS <- ggplot(nmds_data, aes(x = MDS1, y = MDS2, fill = Variable)) +
                    geom_point(shape = 21, color="black", size = 5, stroke = 2) +
                    theme_bw() +
                    theme(axis.text = element_text(color = "black", size = 14),
                            axis.title = element_text(color = "black", size = 14),
                            axis.ticks = element_line(size = 0.8, color = "black"), 
                            axis.ticks.length = unit(0.2, "cm")) +
                    labs(x = nMDS1_label, y = nMDS2_label) + 
                    scale_fill_manual(values = color_palette)
}}

spacer <-  ggplot() +
                geom_blank(fill = "white") +  # Set the background color to white
                theme_void()

title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 05 (vOTU tables)", fontface='bold')

if ("Variable" %in% colnames(metadata)) {{
    plot_col_1 <- plot_grid(title, p_conservative_filtered_vOTUs_CheckV, p_conservative_filtered_vOTUs_length, p_taxa_vOTUs_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C'))
    plot_col_2 <- plot_grid(spacer, p_conservative_filtered_vOTUs_nMDS, spacer, rel_heights = c(0.15, 0.8, 0.2),
                            ncol = 1, labels=c('','D', ''))
    combined_plots <- plot_grid(plot_col_1, plot_col_2, ncol = 2)
    print(combined_plots)
}} else {{
combined_plots <- plot_grid(title, p_conservative_filtered_vOTUs_CheckV, p_conservative_filtered_vOTUs_length, p_taxa_vOTUs_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C'))
}}
ggsave("{output_folder}/Summarize_output_plots.pdf", plot = combined_plots, width = 20, height = 12)
    """.format(votu_table_file_path=votu_table_file_path, metadata_file_path=metadata_file_path, output_folder=output_folder)

    # Run the R code
    robjects.r(r_code)

def create_module_05_06_figures(votu_table_file_path, metadata_file_path, functional_annotation_table_file_path, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating figures for Module 05 and Module 06 using R...\033[0m")

    r_code = """
library(ggplot2)
library(dplyr)
library(cowplot)
library(readr)
library(ComplexUpset)
library(stringr)
library(reshape2)
library(vegan)

data = read.table("{votu_table_file_path}", h=T, sep="\t", dec=".")

checkv_data <- data %>%
  count(checkv_quality = checkv_quality) %>%
  mutate(checkv_quality = factor(checkv_quality, levels = c("Complete", "High-quality", "Medium-quality", "Low-quality", "Not-determined"))) %>%
  mutate(Factor = "checkv_quality")

# Create the box plot
p_conservative_filtered_vOTUs_CheckV <- ggplot(checkv_data, aes(x = Factor, y = n, fill = checkv_quality)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.8) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3", "#9D9D9D", "#686868", "#333333"), 
                    limits = c("Not-determined", "Low-quality", "Medium-quality", "High-quality", "Complete"),
                    name="CheckV quality") +
  coord_flip() +
  labs(y = "Number", x = "") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

length_data <- data %>%
  mutate(Length_Category = case_when(
    virus_length < 10000 ~ "<10 kb",
    virus_length >= 10000 & virus_length <= 50000 ~ "10-50 kb",
    TRUE ~ ">50 kb")) %>%
  count(Length_Category) %>%
  mutate(Length_Category = factor(Length_Category, levels = c("<10 kb", "10-50 kb", ">50 kb"))) %>%
  mutate(Factor = "Length")

p_conservative_filtered_vOTUs_length <- ggplot(length_data, aes(x = Length_Category, y = n)) +
  geom_bar(stat="identity", position= position_dodge(preserve = "single"), fill = '#D3D3D3', color = "black", size = 1) + 
  theme_minimal() +
  coord_flip() +
  labs(y = "Number", x = "Genome length") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

data <- data %>%
  mutate(Taxonomy = sapply(strsplit(as.character(taxonomy), ";"), function(x) {{
    level_5 <- ifelse(length(x) >= 5, x[5], tail(x, 1))
    if (level_5 == "Viruses") {{
      return("Unclassified")
    }} else {{
      return(level_5)
    }}
  }}))

taxa_count_data <- data %>%
  count(Taxonomy = data$Taxonomy) %>%
  mutate(Factor = "Taxonomy")
head(taxa_count_data)

# Define the color palette based on the presence of "Unclassified"
if ("Unclassified" %in% taxa_count_data$Taxonomy) {{
  # If "Unclassified" exists, use one less color than the number of unique taxa
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)) - 1)
  # Assign #D3D3D3 to Unclassified
  color_palette <- c(color_palette, "#D3D3D3")
}} else {{
  # If "Unclassified" doesn't exist, use the number of unique taxa as the number of colors
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)))
}}  

# Create the ggplot
p_taxa_vOTUs_counts <- ggplot(taxa_count_data, aes(x = Factor, y = n, fill = Taxonomy)) +
  geom_bar(position = "fill", stat = "identity", color = "black", size = 1) + 
  scale_fill_manual(values = color_palette) +
  labs(x = "", y = "Percentage (%)", fill = "Taxonomy") + 
  theme_minimal() +
  coord_flip() +
  scale_y_continuous(labels = scales::percent_format()) +
  theme(legend.justification = c(0, 0.5), 
        axis.text = element_text(color = "black", size = 14),
        axis.title = element_text(color = "black", size = 14),
        axis.ticks = element_line(size = 0.8, color = "black"),
        axis.ticks.length = unit(0.2, "cm"))

        
metadata = read.table("{metadata_file_path}", h=T, sep="\t", dec=".")

if ("Variable" %in% colnames(metadata)) {{
    metadata_subset <- metadata[, c('Sample', 'Variable')]
    selected_RPKM_columns <- c("virus_id", grep('^RPKM_', names(data), value = TRUE))
    RPKM_data <- data[selected_RPKM_columns]
    colnames(RPKM_data) <- gsub('^RPKM_', '', colnames(RPKM_data))
    RPKM_data_t <- dcast(melt(RPKM_data, id.vars = "virus_id"), variable ~ virus_id)
    RPKM_data_t <- RPKM_data_t %>% rename(Sample = variable)
    RPKM_data_t_metadata <- merge(x = metadata_subset, y = RPKM_data_t, by = "Sample")
    RPKM_data_t_metadata_bio <- RPKM_data_t_metadata[,-c(1,2)]

    MDS <- cmdscale(vegdist(RPKM_data_t_metadata_bio, method = "bray"), k = 3, eig = T, add = T )
    variance_explained <- round(MDS$eig * 100 / sum(MDS$eig), 1)
    variance_explained_first_two <- head(variance_explained, 2)
    nMDS1_label <- paste0("nMDS 1 (", variance_explained_first_two[1], "%)")
    nMDS2_label <- paste0("nMDS 2 (", variance_explained_first_two[2], "%)")
    RPKM_data_t_metadata_bio_dist <- vegdist(RPKM_data_t_metadata_bio, method = "bray")
    RPKM_data_t_metadata_bio_dist_MDS <- metaMDS(RPKM_data_t_metadata_bio_dist, distance = "bray", k = 3, trymax=1000)
    nmds_coordinates <- data.frame(RPKM_data_t_metadata_bio_dist_MDS$points)
    nmds_data <- cbind(nmds_coordinates, Variable = RPKM_data_t_metadata$Variable)

    color_palette <- colorRampPalette(c("#A36B2B", "#C6AA74", "#EDEAC2", "#94B9A7", "#2686A0"))(length(unique(RPKM_data_t_metadata$Variable)))

    p_conservative_filtered_vOTUs_nMDS <- ggplot(nmds_data, aes(x = MDS1, y = MDS2, fill = Variable)) +
                    geom_point(shape = 21, color="black", size = 5, stroke = 2) +
                    theme_bw() +
                    theme(axis.text = element_text(color = "black", size = 14),
                            axis.title = element_text(color = "black", size = 14),
                            axis.ticks = element_line(size = 0.8, color = "black"), 
                            axis.ticks.length = unit(0.2, "cm")) +
                    labs(x = nMDS1_label, y = nMDS2_label) + 
                    scale_fill_manual(values = color_palette)
}}

if ("_ADS_" %in% str_extract("{functional_annotation_table_file_path}", "_ADS_")) {{
    data_M06 <- readr::read_csv("{functional_annotation_table_file_path}", col_types = cols(ADS_Defense_type = col_character()))
    data_M06$PHROGS_Category <- ifelse(is.na(data_M06$PHROGS_Category), "unknown", data_M06$PHROGS_Category)
    data_M06$PFAM_Annotation <- ifelse(is.na(data_M06$PFAM_Annotation), "unknown", data_M06$PFAM_Annotation)
    data_M06$ADS_Defense_type <- ifelse(is.na(data_M06$ADS_Defense_type), "unknown", data_M06$ADS_Defense_type)
    # Initialize lists for each annotation method
    ADS_list <- data.frame(Viral_gene_ID = character(0))
    Pfam_list <- data.frame(Viral_gene_ID = character(0))
    PHROGS_list <- data.frame(Viral_gene_ID = character(0))
    geNomad_list <- data.frame(Viral_gene_ID = character(0))

    # Filter data for each annotation method and populate the lists
    ADS_list <- data_M06 %>%
    filter(ADS_Defense_type != "unknown" & ADS_Defense_type != "Unknown") %>%
    select(Viral_gene_ID)

    # Filter out rows based on PFAM_Annotation column
    Pfam_list <- data_M06 %>%
    filter(!grepl(paste(c("unknown", "Unknown", "Uncharacterized"), collapse = "|"), PFAM_Annotation)) %>%
    select(Viral_gene_ID)

    PHROGS_list <- data_M06 %>%
    filter(PHROGS_Category != "unknown" & PHROGS_Category != "Unknown") %>%
    select(Viral_gene_ID)

    geNomad_list <- data_M06 %>%
    filter(GENOMAD_Annotation != "unknown" & GENOMAD_Annotation != "Unknown") %>%
    select(Viral_gene_ID)

    # Combine all lists into a single data frame
    upset_data <- data.frame(Viral_gene_ID = data_M06$Viral_gene_ID,
                            ADS = as.integer(data_M06$Viral_gene_ID %in% ADS_list$Viral_gene_ID),
                            Pfam = as.integer(data_M06$Viral_gene_ID %in% Pfam_list$Viral_gene_ID),
                            PHROGS = as.integer(data_M06$Viral_gene_ID %in% PHROGS_list$Viral_gene_ID),
                            geNomad = as.integer(data_M06$Viral_gene_ID %in% geNomad_list$Viral_gene_ID))

    p_annotation_upset <- upset(
                            upset_data, 
                            name='Group',
                            width_ratio=0.3,
                            c('ADS', 'Pfam', 'PHROGS', 'geNomad'),
                            queries = list(upset_query(set = 'ADS', fill = '#EFE32A'),
                                            upset_query(set = 'Pfam', fill = '#CB1800'),
                                            upset_query(set = 'PHROGS', fill = '#00A87E'),
                                            upset_query(set = 'geNomad', fill = '#61BCEC')),
                            base_annotations=list('Shared genes between groups'=(intersection_size(counts=T, col = "black", 
                                                                                                    bar_number_threshold=1, width=0.8) +
                                                                                    scale_y_continuous(expand=expansion(mult=c(0, 0.05))) + 
                                                                                    theme(panel.grid=element_blank(),
                                                                                        panel.background = element_rect(fill = 'white'),
                                                                                        axis.text.x=element_blank(),
                                                                                        axis.text.y = element_text(color="black", size=14),
                                                                                        axis.title = element_text(color="black", size=14),
                                                                                        axis.title.x=element_blank(),
                                                                                        axis.line=element_line(colour='black')))),
                            set_sizes=(upset_set_size(geom=geom_bar(width=0.6, col = "black"),
                                                        position = 'right') + 
                                        geom_text(aes(label=after_stat(count)), hjust=-0.1, stat='count') +
                                        ylab("Nb of genes per group") +
                                        expand_limits(y=7000) +
                                        theme(panel.background =element_blank(),
                                                axis.line.x=element_line(colour='black'),
                                                axis.ticks.y=element_blank(),
                                                axis.text.y=element_blank(),
                                                axis.text.x = element_text(color="black", size=14),
                                                axis.title.x = element_text(color="black", size=14),
                                                axis.title.y=element_blank())),
                            theme = list(legend.text.align = NULL),
                            matrix=(intersection_matrix(geom=geom_point(shape='circle filled', size=3.5, stroke=0.45)) + 
                                        ylab("Database") +
                                        theme(panel.background =element_blank(),
                                            axis.line=element_blank(),
                                            axis.ticks=element_blank(),
                                            axis.title=element_text(color="black", size=14),
                                            axis.text.y = element_text(color="black", size=14),
                                            axis.text.x=element_blank())),
                            stripes=upset_stripes(geom=geom_segment(linewidth=12), colors=c('grey95', 'white')),
                            sort_sets='descending',
                            sort_intersections='descending')
}} else {{
    data_M06 <- readr::read_csv("{functional_annotation_table_file_path}")
    data_M06$PHROGS_Category <- ifelse(is.na(data_M06$PHROGS_Category), "unknown", data_M06$PHROGS_Category)
    data_M06$PFAM_Annotation <- ifelse(is.na(data_M06$PFAM_Annotation), "unknown", data_M06$PFAM_Annotation)
    # Initialize lists for each annotation method
    Pfam_list <- data.frame(Viral_gene_ID = character(0))
    PHROGS_list <- data.frame(Viral_gene_ID = character(0))
    geNomad_list <- data.frame(Viral_gene_ID = character(0))

    # Filter out rows based on PFAM_Annotation column
    Pfam_list <- data_M06 %>%
    filter(!grepl(paste(c("unknown", "Unknown", "Uncharacterized"), collapse = "|"), PFAM_Annotation)) %>%
    select(Viral_gene_ID)

    PHROGS_list <- data_M06 %>%
    filter(PHROGS_Category != "unknown" & PHROGS_Category != "Unknown") %>%
    select(Viral_gene_ID)

    geNomad_list <- data_M06 %>%
    filter(GENOMAD_Annotation != "unknown" & GENOMAD_Annotation != "Unknown") %>%
    select(Viral_gene_ID)

    # Combine all lists into a single data frame
    upset_data <- data.frame(Viral_gene_ID = data_M06$Viral_gene_ID,
                            Pfam = as.integer(data_M06$Viral_gene_ID %in% Pfam_list$Viral_gene_ID),
                            PHROGS = as.integer(data_M06$Viral_gene_ID %in% PHROGS_list$Viral_gene_ID),
                            geNomad = as.integer(data_M06$Viral_gene_ID %in% geNomad_list$Viral_gene_ID))

    p_annotation_upset <- upset(
                            upset_data, 
                            name='Group',
                            width_ratio=0.3,
                            c('Pfam', 'PHROGS', 'geNomad'),
                            queries = list(upset_query(set = 'Pfam', fill = '#CB1800'),
                                            upset_query(set = 'PHROGS', fill = '#00A87E'),
                                            upset_query(set = 'geNomad', fill = '#61BCEC')),
                            base_annotations=list('Shared genes between groups'=(intersection_size(counts=T, col = "black", 
                                                                                                    bar_number_threshold=1, width=0.8) +
                                                                                    scale_y_continuous(expand=expansion(mult=c(0, 0.05))) + 
                                                                                    theme(panel.grid=element_blank(),
                                                                                        panel.background = element_rect(fill = 'white'),
                                                                                        axis.text.x=element_blank(),
                                                                                        axis.text.y = element_text(color="black", size=14),
                                                                                        axis.title = element_text(color="black", size=14),
                                                                                        axis.title.x=element_blank(),
                                                                                        axis.line=element_line(colour='black')))),
                            set_sizes=(upset_set_size(geom=geom_bar(width=0.6, col = "black"),
                                                        position = 'right') + 
                                        geom_text(aes(label=after_stat(count)), hjust=-0.1, stat='count') +
                                        ylab("Nb of genes per group") +
                                        theme(panel.background =element_blank(),
                                                axis.line.x=element_line(colour='black'),
                                                axis.ticks.y=element_blank(),
                                                axis.text.y=element_blank(),
                                                axis.text.x = element_text(color="black", size=14),
                                                axis.title.x = element_text(color="black", size=14),
                                                axis.title.y=element_blank())),
                            theme = list(legend.text.align = NULL),
                            matrix=(intersection_matrix(geom=geom_point(shape='circle filled', size=3.5, stroke=0.45)) + 
                                        ylab("Database") +
                                        theme(panel.background =element_blank(),
                                            axis.line=element_blank(),
                                            axis.ticks=element_blank(),
                                            axis.title=element_text(color="black", size=14),
                                            axis.text.y = element_text(color="black", size=14),
                                            axis.text.x=element_blank())),
                            stripes=upset_stripes(geom=geom_segment(linewidth=12), colors=c('grey95', 'white')),
                            sort_sets='descending',
                            sort_intersections='descending')
}}

spacer <-  ggplot() +
                geom_blank(fill = "white") +  # Set the background color to white
                theme_void()

title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 05 (vOTU tables)and 06 (Functionnal annotation)", fontface='bold')

if ("Variable" %in% colnames(metadata)) {{
    plot_col_1 <- plot_grid(title, p_conservative_filtered_vOTUs_CheckV, p_conservative_filtered_vOTUs_length, p_taxa_vOTUs_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C'))
    plot_col_2 <- plot_grid(spacer, p_conservative_filtered_vOTUs_nMDS, p_annotation_upset, rel_heights = c(0.2, 0.8, 0.8),
                            ncol = 1, labels=c('','D','E'))
    combined_plots <- plot_grid(plot_col_1, plot_col_2, ncol = 2)
}} else {{
    plot_col_1 <- plot_grid(title, p_conservative_filtered_vOTUs_CheckV, p_conservative_filtered_vOTUs_length, p_taxa_vOTUs_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C'))
    plot_col_2 <- plot_grid(spacer,  p_annotation_upset, spacer,  rel_heights = c(0.15, 0.8, 0.2),
                            ncol = 1, labels=c('','D',''))
    combined_plots <- plot_grid(plot_col_1, plot_col_2, ncol = 2)
}}
ggsave("{output_folder}/Summarize_output_plots.pdf", plot = combined_plots, width = 20, height = 12)
    """.format(votu_table_file_path=votu_table_file_path, metadata_file_path=metadata_file_path, functional_annotation_table_file_path=functional_annotation_table_file_path, output_folder=output_folder)

    # Run the R code
    robjects.r(r_code)

def create_module_07_figures(votu_table_file_path, metadata_file_path, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating figures for Module 07 using R...\033[0m")

    # Define R code
    r_code = """
library(ggplot2)
library(dplyr)
library(cowplot)
library(readr)
library(ComplexUpset)
library(stringr)
library(reshape2)
library(vegan)

data = read.table("{votu_table_file_path}", h=T, sep="\t", dec=".")

# Add Variable_Type column to vBins_count_data_M07
vBins_count_data <- data %>%
                    count(Type = data$Type) %>%
                    mutate(x = "Type")
vBins_count_data$Type <- factor(vBins_count_data$Type, levels = c("vBin", "Unbinned_Contig"),
                                    labels = c("vBin", "Unbinned\ncontig"))

p_filtered_vBins <- ggplot(vBins_count_data, aes(x = x, y = n, fill = Type)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.5) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3"), limits = c("Unbinned\ncontig", "vBin")) +
  coord_flip() +
  labs(x = "", y = "Number") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

checkv_data <- data %>%
  group_by(Type) %>%
  count(checkv_quality = checkv_quality) %>%
  mutate(checkv_quality = factor(checkv_quality, levels = c("Complete", "High-quality", "Medium-quality", "Low-quality"))) %>%
  mutate(Factor = "checkv_quality")
checkv_data$Type <- factor(checkv_data$Type, levels = c("vBin", "Unbinned_Contig"),
                                labels = c("vBin", "Unbinned\ncontig"))

# Create the box plot
p_conservative_filtered_vBins_CheckV <- ggplot(checkv_data, aes(x = Type, y = n, fill = checkv_quality)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.8) + 
  theme_minimal() +
  scale_fill_manual(values = c("#D3D3D3", "#9D9D9D", "#686868", "#333333"), 
                    limits = c("Low-quality", "Medium-quality", "High-quality", "Complete"),
                    name="CheckV quality") +
  coord_flip() +
  labs(y = "Number", x = "") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

length_data <- data %>%
  group_by(Type) %>%
  mutate(Length_Category = case_when(
    virus_length < 10000 ~ "<10 kb",
    virus_length >= 10000 & virus_length <= 50000 ~ "10-50 kb",
    TRUE ~ ">50 kb")) %>%
  count(Length_Category) %>%
  mutate(Length_Category = factor(Length_Category, levels = c("<10 kb", "10-50 kb", ">50 kb"))) %>%
  mutate(Factor = "Length")
length_data$Type <- factor(length_data$Type, levels = c("vBin", "Unbinned_Contig"),
                           labels = c("vBin", "Unbinned\ncontig"))

p_conservative_filtered_vBins_length <- ggplot(length_data, aes(x = Length_Category, y = n, fill = Type)) +
  geom_bar(stat="identity", position= position_dodge(preserve = "single"), color = "black", size = 1) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3"), limits = c("Unbinned\ncontig", "vBin")) +
  coord_flip() +
  labs(y = "Number", x = "Genome length") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

data <- data %>%
  mutate(Taxonomy = sapply(strsplit(as.character(taxonomy), ";"), function(x) {{
    level_5 <- ifelse(length(x) >= 5, x[5], tail(x, 1))
    if (level_5 == "Viruses") {{
      return("Unclassified")
    }} else {{
      return(level_5)
    }}
  }}))

taxa_count_data <- data %>%
  count(Taxonomy = data$Taxonomy) %>%
  mutate(Factor = "Taxonomy")
head(taxa_count_data)

# Define the color palette based on the presence of "Unclassified"
if ("Unclassified" %in% taxa_count_data$Taxonomy) {{
  # If "Unclassified" exists, use one less color than the number of unique taxa
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)) - 1)
  # Assign #D3D3D3 to Unclassified
  color_palette <- c(color_palette, "#D3D3D3")
}} else {{
  # If "Unclassified" doesn't exist, use the number of unique taxa as the number of colors
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)))
}}  

# Create the ggplot
p_taxa_vBins_counts <- ggplot(taxa_count_data, aes(x = Factor, y = n, fill = Taxonomy)) +
  geom_bar(position = "fill", stat = "identity", color = "black", size = 1) + 
  scale_fill_manual(values = color_palette) +
  labs(x = "", y = "Percentage (%)", fill = "Taxonomy") + 
  theme_minimal() +
  coord_flip() +
  scale_y_continuous(labels = scales::percent_format()) +
  theme(legend.justification = c(0, 0.5), 
        axis.text = element_text(color = "black", size = 14),
        axis.title = element_text(color = "black", size = 14),
        axis.ticks = element_line(size = 0.8, color = "black"),
        axis.ticks.length = unit(0.2, "cm"))

        
metadata = read.table("{metadata_file_path}", h=T, sep="\t", dec=".")

if ("Variable" %in% colnames(metadata)) {{
    metadata_subset <- metadata[, c('Sample', 'Variable')]
    selected_RPKM_columns <- c("virus_id", grep('^RPKM_', names(data), value = TRUE))
    RPKM_data <- data[selected_RPKM_columns]
    colnames(RPKM_data) <- gsub('^RPKM_', '', colnames(RPKM_data))
    RPKM_data_t <- dcast(melt(RPKM_data, id.vars = "virus_id"), variable ~ virus_id)
    RPKM_data_t <- RPKM_data_t %>% rename(Sample = variable)
    RPKM_data_t_metadata <- merge(x = metadata_subset, y = RPKM_data_t, by = "Sample")
    RPKM_data_t_metadata_bio <- RPKM_data_t_metadata[,-c(1,2)]

    MDS <- cmdscale(vegdist(RPKM_data_t_metadata_bio, method = "bray"), k = 3, eig = T, add = T )
    variance_explained <- round(MDS$eig * 100 / sum(MDS$eig), 1)
    variance_explained_first_two <- head(variance_explained, 2)
    nMDS1_label <- paste0("nMDS 1 (", variance_explained_first_two[1], "%)")
    nMDS2_label <- paste0("nMDS 2 (", variance_explained_first_two[2], "%)")
    RPKM_data_t_metadata_bio_dist <- vegdist(RPKM_data_t_metadata_bio, method = "bray")
    RPKM_data_t_metadata_bio_dist_MDS <- metaMDS(RPKM_data_t_metadata_bio_dist, distance = "bray", k = 3, trymax=1000)
    nmds_coordinates <- data.frame(RPKM_data_t_metadata_bio_dist_MDS$points)
    nmds_data <- cbind(nmds_coordinates, Variable = RPKM_data_t_metadata$Variable)

    color_palette <- colorRampPalette(c("#A36B2B", "#C6AA74", "#EDEAC2", "#94B9A7", "#2686A0"))(length(unique(RPKM_data_t_metadata$Variable)))

    p_conservative_filtered_vBins_nMDS <- ggplot(nmds_data, aes(x = MDS1, y = MDS2, fill = Variable)) +
                    geom_point(shape = 21, color="black", size = 5, stroke = 2) +
                    theme_bw() +
                    theme(axis.text = element_text(color = "black", size = 14),
                            axis.title = element_text(color = "black", size = 14),
                            axis.ticks = element_line(size = 0.8, color = "black"), 
                            axis.ticks.length = unit(0.2, "cm")) +
                    labs(x = nMDS1_label, y = nMDS2_label) + 
                    scale_fill_manual(values = color_palette)


}} else {{
    # Continue execution if metadata does not have a column named "Variable"
    next
}}

if ("Variable" %in% colnames(metadata)) {{
    spacer <-  ggplot() +
                geom_blank(fill = "white") +  # Set the background color to white
                theme_void()

    title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 07 (Binning)", fontface='bold')
    plot_col_1 <- plot_grid(title, p_filtered_vBins, p_conservative_filtered_vBins_CheckV, p_conservative_filtered_vBins_length, p_taxa_vBins_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C', 'D'))
    plot_col_2 <- plot_grid(spacer, p_conservative_filtered_vBins_nMDS, spacer, rel_heights = c(0.1, 0.8, 0.3),
                            ncol = 1, labels=c('','E', ''))
    combined_plots <- plot_grid(plot_col_1, plot_col_2, ncol = 2)
    print(combined_plots)
}} else {{
title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 07 (Binning)", fontface='bold')
combined_plots <- plot_grid(title, p_filtered_vBins, p_conservative_filtered_vBins_CheckV, p_conservative_filtered_vBins_length, p_taxa_vBins_counts,
                            ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C', 'D'))
}}

ggsave("{output_folder}/Summarize_output_plots.pdf", plot = combined_plots, width = 20, height = 12)
    """.format(votu_table_file_path=votu_table_file_path, metadata_file_path=metadata_file_path, output_folder=output_folder)

    # Run the R code
    robjects.r(r_code)

def create_module_06_07_figures(votu_table_file_path, functional_annotation_table_file_path, metadata_file_path, output_folder):
    print(f"\n\033[1m{step_counter.print_main_step()}: Generating figures for Module 06 and Module 07 using R...\033[0m")

    # Define R code
    r_code = """
sink("/dev/null")
library(ggplot2)
library(dplyr)
library(cowplot)
library(readr)
library(ComplexUpset)
library(stringr)
library(reshape2)
library(vegan)

data = read.table("{votu_table_file_path}", h=T, sep="\t", dec=".")

# Add Variable_Type column to vBins_count_data_M07
vBins_count_data <- data %>%
                    count(Type = data$Type) %>%
                    mutate(x = "Type")
vBins_count_data$Type <- factor(vBins_count_data$Type, levels = c("vBin", "Unbinned_Contig"),
                                    labels = c("vBin", "Unbinned\ncontig"))

p_filtered_vBins <- ggplot(vBins_count_data, aes(x = x, y = n, fill = Type)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.5) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3"), limits = c("Unbinned\ncontig", "vBin")) +
  coord_flip() +
  labs(x = "", y = "Number") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

checkv_data <- data %>%
  group_by(Type) %>%
  count(checkv_quality = checkv_quality) %>%
  mutate(checkv_quality = factor(checkv_quality, levels = c("Complete", "High-quality", "Medium-quality", "Low-quality"))) %>%
  mutate(Factor = "checkv_quality")
checkv_data$Type <- factor(checkv_data$Type, levels = c("vBin", "Unbinned_Contig"),
                                labels = c("vBin", "Unbinned\ncontig"))

# Create the box plot
p_conservative_filtered_vBins_CheckV <- ggplot(checkv_data, aes(x = Type, y = n, fill = checkv_quality)) +
  geom_bar(stat="identity", color = "black", size = 1, width = 0.8) + 
  theme_minimal() +
  scale_fill_manual(values = c("#D3D3D3", "#9D9D9D", "#686868", "#333333"), 
                    limits = c("Low-quality", "Medium-quality", "High-quality", "Complete"),
                    name="CheckV quality") +
  coord_flip() +
  labs(y = "Number", x = "") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

length_data <- data %>%
  group_by(Type) %>%
  mutate(Length_Category = case_when(
    virus_length < 10000 ~ "<10 kb",
    virus_length >= 10000 & virus_length <= 50000 ~ "10-50 kb",
    TRUE ~ ">50 kb")) %>%
  count(Length_Category) %>%
  mutate(Length_Category = factor(Length_Category, levels = c("<10 kb", "10-50 kb", ">50 kb"))) %>%
  mutate(Factor = "Length")
length_data$Type <- factor(length_data$Type, levels = c("vBin", "Unbinned_Contig"),
                           labels = c("vBin", "Unbinned\ncontig"))

p_conservative_filtered_vBins_length <- ggplot(length_data, aes(x = Length_Category, y = n, fill = Type)) +
  geom_bar(stat="identity", position= position_dodge(preserve = "single"), color = "black", size = 1) + 
  theme_minimal() +
  scale_fill_manual(values = c("white", "#D3D3D3"), limits = c("Unbinned\ncontig", "vBin")) +
  coord_flip() +
  labs(y = "Number", x = "Genome length") + 
  theme(legend.justification = c(0,0.5), 
        axis.text = element_text(color="black", size=14),
        axis.title = element_text(color="black", size=14),
        axis.ticks = element_line(size = 0.8, color="black"), axis.ticks.length = unit(0.2, "cm"))

data <- data %>%
  mutate(Taxonomy = sapply(strsplit(as.character(taxonomy), ";"), function(x) {{
    level_5 <- ifelse(length(x) >= 5, x[5], tail(x, 1))
    if (level_5 == "Viruses") {{
      return("Unclassified")
    }} else {{
      return(level_5)
    }}
  }}))

taxa_count_data <- data %>%
  count(Taxonomy = data$Taxonomy) %>%
  mutate(Factor = "Taxonomy")
head(taxa_count_data)

# Define the color palette based on the presence of "Unclassified"
if ("Unclassified" %in% taxa_count_data$Taxonomy) {{
  # If "Unclassified" exists, use one less color than the number of unique taxa
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)) - 1)
  # Assign #D3D3D3 to Unclassified
  color_palette <- c(color_palette, "#D3D3D3")
}} else {{
  # If "Unclassified" doesn't exist, use the number of unique taxa as the number of colors
  color_palette <- colorRampPalette(c("#3B9AB1", "#78B7C5", "#EBCC2A", "#E1AE01", "#F22300"))(length(unique(taxa_count_data$Taxonomy)))
}}  

# Create the ggplot
p_taxa_vBins_counts <- ggplot(taxa_count_data, aes(x = Factor, y = n, fill = Taxonomy)) +
  geom_bar(position = "fill", stat = "identity", color = "black", size = 1) + 
  scale_fill_manual(values = color_palette) +
  labs(x = "", y = "Percentage (%)", fill = "Taxonomy") + 
  theme_minimal() +
  coord_flip() +
  scale_y_continuous(labels = scales::percent_format()) +
  theme(legend.justification = c(0, 0.5), 
        axis.text = element_text(color = "black", size = 14),
        axis.title = element_text(color = "black", size = 14),
        axis.ticks = element_line(size = 0.8, color = "black"),
        axis.ticks.length = unit(0.2, "cm"))

        
metadata = read.table("{metadata_file_path}", h=T, sep="\t", dec=".")

if ("Variable" %in% colnames(metadata)) {{
    metadata_subset <- metadata[, c('Sample', 'Variable')]
    selected_RPKM_columns <- c("virus_id", grep('^RPKM_', names(data), value = TRUE))
    RPKM_data <- data[selected_RPKM_columns]
    colnames(RPKM_data) <- gsub('^RPKM_', '', colnames(RPKM_data))
    RPKM_data_t <- dcast(melt(RPKM_data, id.vars = "virus_id"), variable ~ virus_id)
    RPKM_data_t <- RPKM_data_t %>% rename(Sample = variable)
    RPKM_data_t_metadata <- merge(x = metadata_subset, y = RPKM_data_t, by = "Sample")
    RPKM_data_t_metadata_bio <- RPKM_data_t_metadata[,-c(1,2)]

    MDS <- cmdscale(vegdist(RPKM_data_t_metadata_bio, method = "bray"), k = 3, eig = T, add = T )
    variance_explained <- round(MDS$eig * 100 / sum(MDS$eig), 1)
    variance_explained_first_two <- head(variance_explained, 2)
    nMDS1_label <- paste0("nMDS 1 (", variance_explained_first_two[1], "%)")
    nMDS2_label <- paste0("nMDS 2 (", variance_explained_first_two[2], "%)")
    RPKM_data_t_metadata_bio_dist <- vegdist(RPKM_data_t_metadata_bio, method = "bray")
    RPKM_data_t_metadata_bio_dist_MDS <- metaMDS(RPKM_data_t_metadata_bio_dist, distance = "bray", k = 3, trymax=1000)
    nmds_coordinates <- data.frame(RPKM_data_t_metadata_bio_dist_MDS$points)
    nmds_data <- cbind(nmds_coordinates, Variable = RPKM_data_t_metadata$Variable)

    color_palette <- colorRampPalette(c("#A36B2B", "#C6AA74", "#EDEAC2", "#94B9A7", "#2686A0"))(length(unique(RPKM_data_t_metadata$Variable)))

    p_conservative_filtered_vBins_nMDS <- ggplot(nmds_data, aes(x = MDS1, y = MDS2, fill = Variable)) +
                    geom_point(shape = 21, color="black", size = 5, stroke = 2) +
                    theme_bw() +
                    theme(axis.text = element_text(color = "black", size = 14),
                            axis.title = element_text(color = "black", size = 14),
                            axis.ticks = element_line(size = 0.8, color = "black"), 
                            axis.ticks.length = unit(0.2, "cm")) +
                    labs(x = nMDS1_label, y = nMDS2_label) + 
                    scale_fill_manual(values = color_palette)


}}

if ("_ADS_" %in% str_extract("{functional_annotation_table_file_path}", "_ADS_")) {{
    data_M06 <- readr::read_csv("{functional_annotation_table_file_path}", col_types = cols(ADS_Defense_type = col_character()))
    data_M06$PHROGS_Category <- ifelse(is.na(data_M06$PHROGS_Category), "unknown", data_M06$PHROGS_Category)
    data_M06$PFAM_Annotation <- ifelse(is.na(data_M06$PFAM_Annotation), "unknown", data_M06$PFAM_Annotation)
    data_M06$ADS_Defense_type <- ifelse(is.na(data_M06$ADS_Defense_type), "unknown", data_M06$ADS_Defense_type)
    # Initialize lists for each annotation method
    ADS_list <- data.frame(Viral_gene_ID = character(0))
    Pfam_list <- data.frame(Viral_gene_ID = character(0))
    PHROGS_list <- data.frame(Viral_gene_ID = character(0))
    geNomad_list <- data.frame(Viral_gene_ID = character(0))

    # Filter data for each annotation method and populate the lists
    ADS_list <- data_M06 %>%
    filter(ADS_Defense_type != "unknown" & ADS_Defense_type != "Unknown") %>%
    select(Viral_gene_ID)

    # Filter out rows based on PFAM_Annotation column
    Pfam_list <- data_M06 %>%
    filter(!grepl(paste(c("unknown", "Unknown", "Uncharacterized"), collapse = "|"), PFAM_Annotation)) %>%
    select(Viral_gene_ID)

    PHROGS_list <- data_M06 %>%
    filter(PHROGS_Category != "unknown" & PHROGS_Category != "Unknown") %>%
    select(Viral_gene_ID)

    geNomad_list <- data_M06 %>%
    filter(GENOMAD_Annotation != "unknown" & GENOMAD_Annotation != "Unknown") %>%
    select(Viral_gene_ID)

    # Combine all lists into a single data frame
    upset_data <- data.frame(Viral_gene_ID = data_M06$Viral_gene_ID,
                            ADS = as.integer(data_M06$Viral_gene_ID %in% ADS_list$Viral_gene_ID),
                            Pfam = as.integer(data_M06$Viral_gene_ID %in% Pfam_list$Viral_gene_ID),
                            PHROGS = as.integer(data_M06$Viral_gene_ID %in% PHROGS_list$Viral_gene_ID),
                            geNomad = as.integer(data_M06$Viral_gene_ID %in% geNomad_list$Viral_gene_ID))

    p_annotation_upset <- upset(
                            upset_data, 
                            name='Group',
                            width_ratio=0.3,
                            c('ADS', 'Pfam', 'PHROGS', 'geNomad'),
                            queries = list(upset_query(set = 'ADS', fill = '#EFE32A'),
                                            upset_query(set = 'Pfam', fill = '#CB1800'),
                                            upset_query(set = 'PHROGS', fill = '#00A87E'),
                                            upset_query(set = 'geNomad', fill = '#61BCEC')),
                            base_annotations=list('Shared genes between groups'=(intersection_size(counts=T, col = "black", 
                                                                                                    bar_number_threshold=1, width=0.8) +
                                                                                    scale_y_continuous(expand=expansion(mult=c(0, 0.05))) + 
                                                                                    theme(panel.grid=element_blank(),
                                                                                        panel.background = element_rect(fill = 'white'),
                                                                                        axis.text.x=element_blank(),
                                                                                        axis.text.y = element_text(color="black", size=14),
                                                                                        axis.title = element_text(color="black", size=14),
                                                                                        axis.title.x=element_blank(),
                                                                                        axis.line=element_line(colour='black')))),
                            set_sizes=(upset_set_size(geom=geom_bar(width=0.6, col = "black"),
                                                        position = 'right') + 
                                        geom_text(aes(label=after_stat(count)), hjust=-0.1, stat='count') +
                                        ylab("Nb of genes per group") +
                                        expand_limits(y=7000) +
                                        theme(panel.background =element_blank(),
                                                axis.line.x=element_line(colour='black'),
                                                axis.ticks.y=element_blank(),
                                                axis.text.y=element_blank(),
                                                axis.text.x = element_text(color="black", size=14),
                                                axis.title.x = element_text(color="black", size=14),
                                                axis.title.y=element_blank())),
                            theme = list(legend.text.align = NULL),
                            matrix=(intersection_matrix(geom=geom_point(shape='circle filled', size=3.5, stroke=0.45)) + 
                                        ylab("Database") +
                                        theme(panel.background =element_blank(),
                                            axis.line=element_blank(),
                                            axis.ticks=element_blank(),
                                            axis.title=element_text(color="black", size=14),
                                            axis.text.y = element_text(color="black", size=14),
                                            axis.text.x=element_blank())),
                            stripes=upset_stripes(geom=geom_segment(linewidth=12), colors=c('grey95', 'white')),
                            sort_sets='descending',
                            sort_intersections='descending')
}} else {{
    data_M06 <- readr::read_csv("{functional_annotation_table_file_path}")
    data_M06$PHROGS_Category <- ifelse(is.na(data_M06$PHROGS_Category), "unknown", data_M06$PHROGS_Category)
    data_M06$PFAM_Annotation <- ifelse(is.na(data_M06$PFAM_Annotation), "unknown", data_M06$PFAM_Annotation)
    # Initialize lists for each annotation method
    Pfam_list <- data.frame(Viral_gene_ID = character(0))
    PHROGS_list <- data.frame(Viral_gene_ID = character(0))
    geNomad_list <- data.frame(Viral_gene_ID = character(0))

    # Filter out rows based on PFAM_Annotation column
    Pfam_list <- data_M06 %>%
    filter(!grepl(paste(c("unknown", "Unknown", "Uncharacterized"), collapse = "|"), PFAM_Annotation)) %>%
    select(Viral_gene_ID)

    PHROGS_list <- data_M06 %>%
    filter(PHROGS_Category != "unknown" & PHROGS_Category != "Unknown") %>%
    select(Viral_gene_ID)

    geNomad_list <- data_M06 %>%
    filter(GENOMAD_Annotation != "unknown" & GENOMAD_Annotation != "Unknown") %>%
    select(Viral_gene_ID)

    # Combine all lists into a single data frame
    upset_data <- data.frame(Viral_gene_ID = data_M06$Viral_gene_ID,
                            Pfam = as.integer(data_M06$Viral_gene_ID %in% Pfam_list$Viral_gene_ID),
                            PHROGS = as.integer(data_M06$Viral_gene_ID %in% PHROGS_list$Viral_gene_ID),
                            geNomad = as.integer(data_M06$Viral_gene_ID %in% geNomad_list$Viral_gene_ID))

    p_annotation_upset <- upset(
                            upset_data, 
                            name='Group',
                            width_ratio=0.3,
                            c('Pfam', 'PHROGS', 'geNomad'),
                            queries = list(upset_query(set = 'Pfam', fill = '#CB1800'),
                                            upset_query(set = 'PHROGS', fill = '#00A87E'),
                                            upset_query(set = 'geNomad', fill = '#61BCEC')),
                            base_annotations=list('Shared genes between groups'=(intersection_size(counts=T, col = "black", 
                                                                                                    bar_number_threshold=1, width=0.8) +
                                                                                    scale_y_continuous(expand=expansion(mult=c(0, 0.05))) + 
                                                                                    theme(panel.grid=element_blank(),
                                                                                        panel.background = element_rect(fill = 'white'),
                                                                                        axis.text.x=element_blank(),
                                                                                        axis.text.y = element_text(color="black", size=14),
                                                                                        axis.title = element_text(color="black", size=14),
                                                                                        axis.title.x=element_blank(),
                                                                                        axis.line=element_line(colour='black')))),
                            set_sizes=(upset_set_size(geom=geom_bar(width=0.6, col = "black"),
                                                        position = 'right') + 
                                        geom_text(aes(label=after_stat(count)), hjust=-0.1, stat='count') +
                                        ylab("Nb of genes per group") +
                                        theme(panel.background =element_blank(),
                                                axis.line.x=element_line(colour='black'),
                                                axis.ticks.y=element_blank(),
                                                axis.text.y=element_blank(),
                                                axis.text.x = element_text(color="black", size=14),
                                                axis.title.x = element_text(color="black", size=14),
                                                axis.title.y=element_blank())),
                            theme = list(legend.text.align = NULL),
                            matrix=(intersection_matrix(geom=geom_point(shape='circle filled', size=3.5, stroke=0.45)) + 
                                        ylab("Database") +
                                        theme(panel.background =element_blank(),
                                            axis.line=element_blank(),
                                            axis.ticks=element_blank(),
                                            axis.title=element_text(color="black", size=14),
                                            axis.text.y = element_text(color="black", size=14),
                                            axis.text.x=element_blank())),
                            stripes=upset_stripes(geom=geom_segment(linewidth=12), colors=c('grey95', 'white')),
                            sort_sets='descending',
                            sort_intersections='descending')
}}

spacer <-  ggplot() +
            geom_blank(fill = "white") +  # Set the background color to white
            theme_void()

title <- ggdraw() + draw_label("Summarized plots generated by MVP\nModule 06 (Functionnal annotation) and 07 (Binning)", fontface='bold')

if ("Variable" %in% colnames(metadata)) {{
    plot_col_1 <- plot_grid(title, p_filtered_vBins, p_conservative_filtered_vBins_CheckV, p_conservative_filtered_vBins_length, p_taxa_vBins_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C', 'D'))
    plot_col_2 <- plot_grid(spacer, p_conservative_filtered_vBins_nMDS, p_annotation_upset, rel_heights = c(0.2, 0.8, 0.8),
                            ncol = 1, labels=c('','E','F'))
    combined_plots <- plot_grid(plot_col_1, plot_col_2, ncol = 2)
    print(combined_plots)
}} else {{
    plot_col_1 <- plot_grid(title, p_filtered_vBins, p_conservative_filtered_vBins_CheckV, p_conservative_filtered_vBins_length, p_taxa_vBins_counts,
                                ncol = 1, align = "hv", rel_heights = c(0.3, 0.6, 0.6, 1, 0.5), labels=c('','A','B', 'C', 'D'))
    plot_col_2 <- plot_grid(spacer, p_annotation_upset, spacer, ncol = 1, rel_heights = c(0.1, 0.6, 0.3), labels=c('', 'E',''))
    combined_plots <- plot_grid(plot_col_1, plot_col_2, rel_widths = c(0.4, 0.6), ncol = 2)
}}

ggsave("{output_folder}/Summarize_output_plots.pdf", plot = combined_plots, width = 20, height = 12)
    """.format(votu_table_file_path=votu_table_file_path, functional_annotation_table_file_path=functional_annotation_table_file_path, metadata_file_path=metadata_file_path, output_folder=output_folder)

    # Run the R code
    robjects.r(r_code)

def main(args):
    # Capture start time
    start_time = datetime.datetime.now()
    formatted_start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {formatted_start_time}\n") 

    metadata_file_path = os.path.join(args['metadata'])

    # Find the maximum module ran
    max_module, max_module_folder = find_max_module(args)

    # Find all summary reports
    summary_reports_dict, num_summary_reports = find_summary_reports(args, max_module_folder)

    # Get the number of summary reports in the '01_GENOMAD' directory

    print(f"\nYou have processed {num_summary_reports} samples through MVP until Module 0{max_module} ({max_module_folder}).\n")

    output_folder, output_filename = create_output_folder(args)

    # Generate the summary report
    generate_summary_report(output_filename, max_module, max_module_folder, num_summary_reports, summary_reports_dict)

    # Copy additional files based on max_module_folder

    copy_additional_files(args, max_module_folder, output_folder)

    # List of R packages to install
    install_r_packages()

    votu_table_file_name, votu_table_file_path, functional_annotation_table_file_path = find_R_tables(output_folder)

    # Run R code
    if votu_table_file_name.startswith('MVP_07_'):
        if functional_annotation_table_file_path is not None:
            create_module_06_07_figures(votu_table_file_path, functional_annotation_table_file_path, metadata_file_path, output_folder)
        else:
            create_module_07_figures(votu_table_file_path, metadata_file_path, output_folder)
    elif votu_table_file_name.startswith('MVP_05_'):
        if functional_annotation_table_file_path is not None:
            create_module_05_06_figures(votu_table_file_path, metadata_file_path, functional_annotation_table_file_path, output_folder)
        else:
            create_module_05_figures(votu_table_file_path, metadata_file_path, output_folder)
    elif votu_table_file_name.startswith('MVP_03_'):
        if functional_annotation_table_file_path is not None:
            create_module_03_06_figures(votu_table_file_path, metadata_file_path, functional_annotation_table_file_path, output_folder)
        else:
            create_module_03_figures(votu_table_file_path, metadata_file_path, output_folder)

    # Capture end time
    end_time = datetime.datetime.now()
    formatted_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")

    # Calculate duration
    duration = end_time - start_time

    message1 = f"\033[1mModule 100 finished: final summary report, and summary figures generated in {output_folder}!\033[0m\n"
    message2 = "\033[1mWe hope you enjoyed using MVP script, and you can now explore your data!\033[0m\n"
    message3 = "\n\033[1mPlease don't forget to cite MVP!\033[0m"
    line_of_stars = '*' * len(message1)
    print()
    print(line_of_stars)
    print(message1)
    print(f"Start Time: {formatted_start_time}") 
    print(f"End Time: {formatted_end_time}")
    print(f"Running Time: {duration.total_seconds():.2f} seconds\n")
    print(message2)
    print(message3)
    print(line_of_stars)
    print()


import argparse
import sys
import mvip

def cli():
    parser = argparse.ArgumentParser(
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=f"""MVP v{mvip.__version__}: assessing the quality of metagenome-assembled viral genomes
https://gitlab.com/ccoclet/mvp

usage: mvip <program> [options]

programs:
    MVP_00_set_up_MVP                   Check for any potential errors/issues in the metadata and the sequencing/read files, create all the directories that MVP needs, and install the latest versions of geNomad and checkV databases.
    MVP_01_run_genomad_checkv           Run geNomad and CheckV.
    MVP_02_filter_genomad_checkv        Merge and filter geNomad and CheckV outputs.
    MVP_03_do_clustering                Sequence clustering based on pairwise ANI.
    MVP_04_do_read_mapping              Run CoverM to calculate coverage based on read mapping, using the sorted BAM files sorted by reference, and return to one tabular file per sample.
    MVP_05_create_vOTU_table            Merge all the CoverM output tables and create a set of viral OTU tables based on the cutoffs (i.e., horizontal coverage) and filtration mode (i.e., conservative and relaxed).
    MVP_06_do_functional_annotation     Functional annotation of protein sequences against multiple databases.
    MVP_07_do_binning                   Run vRhyme for binning virus genomes and return outputs.
    MVP_99_prep_MIUViG_submission       Additional module to assist with submitting metagenome-assembled viral genome(s) to GenBank, including MIUViG metadata.
    MVP_100_summarize_outputs           Summarize outputs and create figures.""",
    )

    subparsers = parser.add_subparsers(help=argparse.SUPPRESS)

    MVP_00_set_up_MVP_parser = subparsers.add_parser(
        "MVP_00_set_up_MVP",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Check for any potential errors/issues in the metadata and the sequencing/read files, create all the directories that MVP needs, and install the latest versions of geNomad and checkV databases.
\nusage: mvip MVP_00_set_up_MVP -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_00_set_up_MVP.fetch_arguments(MVP_00_set_up_MVP_parser)

    MVP_01_run_genomad_checkv_parser = subparsers.add_parser(
        "MVP_01_run_genomad_checkv",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Run geNomad and CheckV.
\nusage: mvip MVP_01_run_genomad_checkv -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_01_run_genomad_checkv.fetch_arguments(MVP_01_run_genomad_checkv_parser)

    MVP_02_filter_genomad_checkv_parser = subparsers.add_parser(
        "MVP_02_filter_genomad_checkv",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Merge and filter geNomad and CheckV outputs.
\nusage: mvip MVP_02_filter_genomad_checkv -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_02_filter_genomad_checkv.fetch_arguments(MVP_02_filter_genomad_checkv_parser)

    MVP_03_do_clustering_parser = subparsers.add_parser(
        "MVP_03_do_clustering",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Sequence clustering based on pairwise ANI.
\nusage: mvip MVP_03_do_clustering -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_03_do_clustering.fetch_arguments(MVP_03_do_clustering_parser)

    MVP_04_do_read_mapping_parser = subparsers.add_parser(
        "MVP_04_do_read_mapping",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Run CoverM to calculate coverage based on read mapping, using the sorted BAM files sorted by reference, and return to one tabular file per sample.
\nusage: mvip MVP_04_do_read_mapping -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_04_do_read_mapping.fetch_arguments(MVP_04_do_read_mapping_parser)

    MVP_05_create_vOTU_table_parser = subparsers.add_parser(
        "MVP_05_create_vOTU_table",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Merge all the CoverM output tables and create a set of viral OTU tables based on the cutoffs (i.e., horizontal coverage) and filtration mode (i.e., conservative and relaxed).
\nusage: mvip MVP_05_create_vOTU_table -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_05_create_vOTU_table.fetch_arguments(MVP_05_create_vOTU_table_parser)

    MVP_06_do_functional_annotation_parser = subparsers.add_parser(
        "MVP_06_do_functional_annotation",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Functional annotation of protein sequences against multiple databases.
\nusage: mvip MVP_06_do_functional_annotation -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_06_do_functional_annotation.fetch_arguments(MVP_06_do_functional_annotation_parser)

    MVP_07_do_binning_parser = subparsers.add_parser(
        "MVP_07_do_binning",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Run vRhyme for binning virus genomes and return outputs.
\nusage: mvip MVP_07_do_binning -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_07_do_binning.fetch_arguments(MVP_07_do_binning_parser)

    MVP_99_prep_MIUViG_submission_parser = subparsers.add_parser(
        "MVP_99_prep_MIUViG_submission",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Additional module to assist with submitting metagenome-assembled viral genome(s) to GenBank, including MIUViG metadata.
\nusage: mvip MVP_99_prep_MIUViG_submission -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_99_prep_MIUViG_submission.fetch_arguments(MVP_99_prep_MIUViG_submission_parser)

    MVP_100_summarize_outputs_parser = subparsers.add_parser(
        "MVP_100_summarize_outputs",
        usage=argparse.SUPPRESS,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Summarize outputs and create figures.
\nusage: mvip MVP_100_summarize_outputs -i <working_directory_path> -m <metadata_path> [options]""",
    )
    mvip.MVP_100_summarize_outputs.fetch_arguments(MVP_100_summarize_outputs_parser)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)
    elif len(sys.argv) == 2:
        if sys.argv[1] == "MVP_00_set_up_MVP":
            MVP_00_set_up_MVP_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_01_run_genomad_checkv":
            MVP_01_run_genomad_checkv_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_02_filter_genomad_checkv":
            MVP_02_filter_genomad_checkv_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_03_do_clustering":
            MVP_03_do_clustering_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_04_do_read_mapping":
            MVP_04_do_read_mapping_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_05_create_vOTU_table":
            MVP_05_create_vOTU_table_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_06_do_functional_annotation":
            MVP_06_do_functional_annotation_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_07_do_binning":
            MVP_07_do_binning_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_99_prep_MIUViG_submission":
            MVP_99_prep_MIUViG_submission_parser.print_help()
            sys.exit(0)
        elif sys.argv[1] == "MVP_100_summarize_outputs":
            MVP_100_summarize_outputs_parser.print_help()
            sys.exit(0)
            
    args = vars(parser.parse_args())
    args["func"](args)


"""
MVP: Modular Viromics Pipeline for studying viruses identified from sequencing data
"""

from mvip.modules import (
    MVP_00_set_up_MVP,
    MVP_01_run_genomad_checkv,
    MVP_02_filter_genomad_checkv,
    MVP_03_do_clustering,
    MVP_04_do_read_mapping,
    MVP_05_create_vOTU_table,
    MVP_06_do_functional_annotation,
    MVP_07_do_binning,
    MVP_99_prep_MIUViG_submission,
    MVP_100_summarize_outputs,
)

__version__ = "1.1.1"


import argparse
import glob
import os
import subprocess
import sys
import csv
import shutil
import subprocess as sp
import datetime
import time
from os.path import getsize
import pandas as pd
from functools import reduce
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import mvip

def fetch_arguments(parser):
    parser.set_defaults(func=main)
    parser.set_defaults(program="MVP_99_prep_MIUViG_submission")
    parser.add_argument(
        "--input", "-i",
        help="Path to your working directory where you want to run MVP.",
    )
    parser.add_argument(
        "--metadata", "-m",
        help="Path to your metadata that you want to use to run MVP.",
    )
    parser.add_argument('-g', '--genome', help='Identifier of the sequence to be processed.', required=True)
    parser.add_argument('-s', '--step', help='Should be one of "setup_metadata" (to be run first) or "prep_submission" (once sequence metadata have been checked and completed).', choices=['setup_metadata', 'prep_submission'], required=True)
    parser.add_argument('-t', '--template', help='path to the BioSample submission template file, generated from https://submit.ncbi.nlm.nih.gov/genbank/template/submission/, only required for the step 2: prep_submission', required=False)


class StepCounter:
    def __init__(self):
        self.current_step = 0
        self.current_substep = 'A'

    def print_main_step(self):
        self.current_step += 1
        self.current_substep = 'A'  # Reset sub-step counter to 'A'
        return f"Step {self.current_step}"

    def print_sub_step(self):
        if self.current_substep == 'A':  # Only increment main step if starting a new sub-step series
            main_step = self.print_main_step()
        else:
            main_step = f"Step {self.current_step}"  # Use current main step without incrementing
        sub_step = self.current_substep
        self.current_substep = chr(ord(self.current_substep) + 1)
        return f"{main_step}{sub_step}"
    
    def reset(self):
        self.current_step = 0
        self.current_substep = 'A'

# Example usage:
step_counter = StepCounter()


# Prep a quick function to load only some specific sequences from a fasta file, may be useful in multiple places
def load_seq(in_fasta,list_ids):
    seq_return = {}
    for seq_record in SeqIO.parse(in_fasta, "fasta"):
        if seq_record.id in list_ids:
            seq_return[seq_record.id] = seq_record.seq

    return seq_return

def create_output_directory(args):
    # prep the output directory if needed
    r_dir = os.path.join(args['input'], '99_GENBANK_SUBMISSION')
    tpl_dir = os.path.join(args['input'], '99_GENBANK_SUBMISSION', 'UViG_metadata_tables/')
    sub_dir = os.path.join(args['input'], '99_GENBANK_SUBMISSION', 'UViG_submission_files/')
    if not os.path.exists(r_dir):
        os.makedirs(r_dir)
    if not os.path.exists(tpl_dir):
        os.makedirs(tpl_dir)
    if not os.path.exists(sub_dir):
        os.makedirs(sub_dir)

    return tpl_dir, sub_dir

def create_template_path():
    # Construct the path to the template files
    script_directory = os.path.dirname(os.path.abspath(__file__))
    # Get the directory of the mvp directory
    mvp_directory = os.path.dirname(script_directory)
    data_directory = os.path.join(mvp_directory, "data")
    template_metadata = os.path.join(data_directory, "miuvig_template-custom.txt")

    return template_metadata

def load_files(args, tpl_dir):
    # Path for output files:
    out_metadata = os.path.join(tpl_dir, args['genome'] + "_metadata.tsv")
    out_annotation = os.path.join(tpl_dir, args['genome'] + "_annotation.tsv")
    # Load sample metadata
    metadata = pd.read_csv(args['metadata'], sep='\t')
    # Load list of sequences
    summary_file = os.path.join(args['input'], '03_CLUSTERING', 'MVP_03_All_Sample_Unfiltered_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
    seq_summary = pd.read_csv(summary_file, sep='\t')

    return out_metadata, out_annotation, metadata, summary_file, seq_summary

# Prep the functions we need for that
def add_gene(vec,o, seq_summary):
    if vec["strand"] == -1:
        if vec['start'] < 3:
            o.write(f"{vec['end']}\t>1\tCDS\n") ## Right now assuming that a gene ending on the edge is likely to be partial, should be better though because there is a small but not impossible chance of a gene ending right on the edge, we'll see if this causes issues down the road
        else:
            o.write(f"{vec['end']}\t{vec['start']}\tCDS\n")
    else:
        if vec['end'] > (seq_summary["virus_length"]-3):
            o.write(f"{vec['start']}\t>{vec['end']}\tCDS\n")
    o.write(f"\t\t\tcodon_start\t1\n")
    o.write(f"\t\t\tinference\tab initio prediction:Prodigal-gv\n")
    o.write(f"\t\t\tlocus_tag\t{vec['Viral_gene_ID']}\n")
    if vec['GENOMAD_Annotation'] == "Unknown":
        vec['GENOMAD_Annotation'] = "hypothetical protein"
    o.write(f"\t\t\tproduct\t{vec['GENOMAD_Annotation']}\n")

def set_up_metadata(args, seq_summary, summary_file, metadata, template_metadata, out_metadata, out_annotation, tpl_dir):
    print(f"\033[1m{step_counter.print_main_step()}: Selecting the corresponding line, exiting if can't find it...\033[0m")
    seq_summary = seq_summary[seq_summary["virus_id"] == args['genome']]
    if len(seq_summary) != 1:
        if len(seq_summary) == 0:
            sys.exit(f"ERROR - We could not find {args['genome']} in {summary_file}, please double check that the genome id you provided is correct")
        elif len(seq_summary) > 1:
            sys.exit(f"ERROR - We expected to select a single row based on {args['genome']}, but we got more than one, please double check")
    seq_summary = seq_summary.squeeze()
    # Throw a warning if low-quality, not recommended
    if seq_summary["checkv_quality"] == "Low-quality":
        print("")
        print("WARNING - the sequence selected was predicted to be a partial genome of low quality, it is typically not recommended to submit these low-quality sequences to public databases")
        print("Press Enter to continue or q to stop")
        test = input()
        if test == "q" or test == "Q":
            sys.exit()

    print(f"\n\033[1m{step_counter.print_main_step()}: Load the version information for each database and tool...\033[0m")
    version_file = os.path.join(args['input'], "MVP_00_Summary_Report.txt")
    info_version = {}
    with open(version_file,"r") as f:
        tag = 0
        for line in f.readlines():
            line = line.strip()
            if line.startswith("Tools and versions:"):
                tag = 1
            elif tag == 1:
                t = line.split(" ")
                info_version[t[0]] = t[1]
            elif line.startswith(" "):
                tag = 0

    print(f"\n\033[1m{step_counter.print_main_step()}: Guessing the sample id for this sequence...\033[0m")
    info_analysis = {}
    for row_index, (_, row) in enumerate(metadata.iterrows(), start=1):
        sample_name = row['Sample']
        test_file = os.path.join(args['input'], '02_CHECK_V', str(sample_name), 'MVP_02_' + str(sample_name) + '_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Quality_Summary.tsv')
        test_df = pd.read_csv(test_file, sep='\t')
        if args['genome'] in test_df["virus_id"].values:
            print(f"genome {args['genome']} was obtained from sample {sample_name}")
            seq_summary["sample"] = sample_name
            ## Add other info if needed
            test_df =  test_df[test_df["virus_id"] == args['genome']].squeeze()
            seq_summary["provirus"] = test_df["provirus"]
            seq_summary["name"] = args['genome']
            tax_tab = test_df["taxonomy"].split(";")
            ## taxonomy -> used for organism and lineage
            if len(tax_tab) > 0:
                seq_summary["organism"] = tax_tab[-1] + " sp. " + args['genome']
            else:
                seq_summary["organism"] = "unclassified virus"
            seq_summary["lineage"] = test_df["taxonomy"]
            seq_summary["virus_length"] = test_df["virus_length"]
            ## genetic code automatically selected by geNomad
            seq_summary["gcode"] = test_df["genetic_code"]
            # Load topology information from geNomad
            step_one_result_one_path = os.path.join(args['input'], '01_GENOMAD', str(sample_name), str(sample_name) + '_Viruses_Genomad_Output', '*_summary', '*_virus_summary.tsv')
            step_one_result_one_file = glob.glob(step_one_result_one_path)
            step_one_result_one_file = step_one_result_one_file[0]
            test_df = pd.read_csv(step_one_result_one_file, sep='\t')
            test_df =  test_df[test_df["seq_name"] == args['genome']].squeeze() ## Should be a scalar now, because it should be a single line per genome
            seq_summary["topology"] = "linear"
            # if test_df["topology"].isin(["DTR"]).any():
            if test_df["topology"] == "DTR":
                seq_summary["topology"] = "circular"
            # Load information about identification
            info_analysis["id_tool"] = "geNomad" ## Setup defaults just in case
            info_analysis["id_version"] = info_version["genomad"]
            info_analysis["check_tool"] = "CheckV"
            info_analysis["check_version"] = info_version["checkv"]
            # Detailed completeness
            comp_file_one = os.path.join(args['input'], '02_CHECK_V', str(sample_name), str(sample_name) + '_Viruses_CheckV_Output', 'completeness.tsv')
            if os.path.exists(comp_file_one):
                comp_df = pd.read_csv(comp_file_one, sep='\t')
                if args['genome'] in comp_df["contig_id"].values:
                    comp_df = comp_df[comp_df["contig_id"] == args['genome']].squeeze()
                    if comp_df["aai_confidence"] == "high" or comp_df["aai_confidence"] == "medium":
                        seq_summary["compl_appr"] = "reference-based"
                        seq_summary["compl_score"] = comp_df["aai_completeness"]
                        seq_summary["estimated_size"] = int(comp_df["aai_expected_length"])
            comp_file_two = os.path.join(args['input'], '02_CHECK_V', str(sample_name), str(sample_name) + '_Proviruses_CheckV_Output', 'completeness.tsv')
            if os.path.exists(comp_file_two):
                comp_df = pd.read_csv(comp_file_one, sep='\t')
                if args['genome'] in comp_df["contig_id"].values:
                    comp_df = comp_df[comp_df["contig_id"] == args['genome']].squeeze()
                    if comp_df["aai_confidence"] == "high" or comp_df["aai_confidence"] == "medium":
                        seq_summary["compl_appr"] = "reference-based"
                        seq_summary["compl_score"] = comp_df["aai_completeness"]
                        seq_summary["estimated_size"] = int(comp_df["aai_expected_length"])

    print(f"\n\033[1m{step_counter.print_main_step()}: Loading information about vOTU clustering, and making sure this sequence is a representative...\033[0m")
    seq_summary["otu_seq_comp_appr"] = info_analysis["check_tool"] + " clustering;" + str(info_analysis["check_version"])
    step_three_report = os.path.join(args['input'], '03_CLUSTERING', 'MVP_03_Summary_Report.txt')
    with open(step_three_report,"r") as f:
        tmp = {}
        for line in f.readlines():
            line = line.strip()
            if line.startswith("03_do_clustering.py"):
                t = line.split(" ")
                for i in range(len(t)):
                    if t[i] == "--min_ani":
                        tmp["ani"] = t[i+1]
                    if t[i] == "--min_tcov":
                        tmp["af"] = t[i+1]
        if tmp["ani"] != "" and tmp["af"] != "":
            seq_summary["otu_class_appr"] = str(tmp["ani"]) + "% ANI;" + str(tmp["af"]) + "% AF; greedy incremental clustering"
    step_three_result = os.path.join(args['input'], '03_CLUSTERING', 'tmp', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences_Clustering_ANI_Clusters.tsv')
    with open(step_three_result,"r") as f:
        tmp = {}
        for line in f.readlines():
            line = line.strip()
            t = line.split("\t")
            list = t[1].split(",")
            if args['genome'] == t[0]:
                seq_summary["repr"] = t[0]
            elif args['genome'] in list:
                print("")
                print(f"WARNING - THE SEQUENCE YOU SELECTED WAS NOT A vOTU REPRESENTATIVE, YOU PROBABLY WANT TO CHANGE THIS AND PICK THE REPRESENTATIVE ({t[0]}) INSTEAD")
                print("Press Enter to continue or q to stop")
                test = input()
                if test == "q" or test == "Q":
                    sys.exit()
                seq_summary["repr"] = t[0]

    print(f"\n\033[1m{step_counter.print_main_step()}: Loading information from coverage...\033[0m")
    step_four_result = os.path.join(args['input'], '04_READ_MAPPING', str(sample_name), str(sample_name) + '_CoverM.tsv')
    if os.path.exists(step_four_result):
        with open(step_four_result,"r") as f:
            for line in f.readlines():
                line = line.strip()
                if t[1] == seq_summary["repr"]:
                    seq_summary["coverage"] = "{:.2f}".format(t[3])
                    # Coverage is used in two places, formatting the one for Assembly-Data
                    seq_summary["Genome Coverage"] = seq_summary["coverage"] + "x"
    else:
        print("")
        print(f"WARNING - No coverage information found for sample {sample_name}, you will need to enter the coverage manually (we recommend looking at column 'Trimmed Mean' in the CoverM output in step 4 for your sample of interest) ")
        print("Press Enter to continue or q to stop")
        test = input()
        if test == "q" or test == "Q":
            sys.exit()
        seq_summary["coverage"] = ""
        seq_summary["Genome Coverage"] = ""

    print(f"\n\033[1m{step_counter.print_main_step()}: Reformatting the field that need it...\033[0m")
    seq_summary["completedness"] = "partial" ## Regardless of contigs topology, guidelines are that completedness for GenBank should be partial until completeness was experimentally verified
    seq_summary["metagenomic"] = "TRUE"
    seq_summary["environmental_sample"] = "TRUE"
    ## detec_type
    if seq_summary["provirus"] == "Yes":
        seq_summary["detec_type"] = "provirus (UpViG)"
        seq_summary["topology"] = "linear"
    else:
        seq_summary["detec_type"] = "independent sequence (UViG)"
    # assembly_qual
    if seq_summary["miuvig_quality"] == "High-quality":
        seq_summary["assembly_qual"] = seq_summary["miuvig_quality"] + " draft genome"
    elif seq_summary["checkvg_quality"] == "Medium-quality":
        seq_summary["assembly_qual"] = seq_summary["Medium-quality"] + " draft genome"
    else:seq_summary["assembly_qual"] = "Low-quality draft genome"
    # number_contig
    seq_summary["number_contig"] = 1 ## Note: could be modified later if we want to accomdate bins, e.g. from vRhyme
    # pred_genome_struc
    seq_summary["pred_genome_struc"] = "non-segmented" ## NOTE - THIS SHOULD BE FINE MOST OF THE TIME, BUT COULD BE REFINED BASED ON TAXONOMY !
    # pred_genome_type
    seq_summary["pred_genome_type"] = seq_summary["Genome type"]
    # softwares used for identification, completeness, and taxonomy
    seq_summary["vir_ident_software"] = info_analysis["id_tool"] + ";" + str(info_analysis["id_version"]) + ";" + str(seq_summary["virus_score"])
    seq_summary["compl_software"] = info_analysis["check_tool"] + ";" + str(info_analysis["check_version"])
    seq_summary["tax_class"] = info_analysis["id_tool"] + ";" + str(info_analysis["id_version"])
    seq_summary["tax_ident"] = "multi-marker approach" ## geNomad, by default a multi-marker approach
    # feat_pred and similar annotation questions
    seq_summary["annot"] = "geNomad" + ";" + info_version["genomad"] ## Features predicted from geNomad
    seq_summary["feat_pred"] = "prodigal-gv;geNomad" + info_version["genomad"] ## Features predicted from geNomad
    seq_summary["ref_db"] = "Pfam" ## Features predicted from geNomad
    seq_summary["sim_search_meth"] = "mmseqs2" + ";" + info_version["mmseqs2"]
    print("We loaded all the data we wanted, now we create the first metadata and annotation files")

    print(f"\n\033[1m{step_counter.print_main_step()}: Read the template and write the output file with genome-specific information...\033[0m")
    with open(template_metadata,"r") as f, open(out_metadata,"w") as o:
        tmp = {}
        freader = csv.reader(f, delimiter='\t', quoting=csv.QUOTE_NONE)
        mwriter = csv.writer(o, delimiter='\t',quotechar='"', quoting=csv.QUOTE_MINIMAL)
        for row in freader:
            if len(row) > 0:
                if row[0].startswith("#"):
                    row[1] = args['genome']
                elif row[0] != "":
                    if row[0] in seq_summary:
                        row[1] = seq_summary[row[0]]
            mwriter.writerow(row)

    print(f"\n\033[1m{step_counter.print_main_step()}: Creating the annotation file we will need based on the gene annotation by geNomad...\033[0m")
    annotation_file = os.path.join(args['input'], '06_FUNCTIONAL_ANNOTATION', 'MVP_06_All_Sample_Filtered_Relaxed_Merged_Genomad_CheckV_Virus_Proviruses_Gene_Annotation_GENOMAD.tsv')
    annot_df = pd.read_csv(annotation_file, sep='\t')
    annot_df =  annot_df[annot_df["Contig_name"] == args['genome']]
    with open(out_annotation,"w") as o:
        o.write(">Feature "+args['genome']+"\n")
        annot_df.apply(lambda x: add_gene(x,o, seq_summary), axis=1)

def prep_submission(args, out_metadata, out_annotation, sub_dir):
    print(f"\n\033[1m{step_counter.print_main_step()}: Checking that all the required files are here...\033[0m")
    out_fasta = os.path.join(sub_dir, args['genome'] + "_genome.fna")
    clean_metadata = os.path.join(sub_dir, args['genome'] + "_comments.cmt")
    final_annotation = os.path.join(sub_dir, args['genome'] + "_genome.tbl")
    template_file = args['template']
    if not(os.path.exists(out_metadata)):
        sys.exit(f"{out_metadata} does not seem to exist, please run step 'setup_metadata' first")
    if not(os.path.exists(out_annotation)):
        sys.exit(f"{out_annotation} does not seem to exist, please run step 'setup_metadata' first")
    if template_file == None or not(os.path.exists(template_file)):
        sys.exit(f"Please provide a path for a template file with the argument '-t' for this submission (see https://submit.ncbi.nlm.nih.gov/genbank/template/submission/)")
    if not(os.path.exists(template_file)):
        sys.exit(f"Template file {template_file} does not seem to exist, please generate a BioSample template file for this submission (see https://submit.ncbi.nlm.nih.gov/genbank/template/submission/), and provide it with the argument '--template'")

    print(f"\n\033[1m{step_counter.print_main_step()}: Loading metadata and check that all the required information is entered...\033[0m")
    tmp_metadata = {}
    with open(out_metadata,"r") as f:
        freader = csv.reader(f, delimiter='\t', quoting=csv.QUOTE_NONE)
        for row in freader:
            if len(row) > 1:
                if row[1] != "":
                    tmp_metadata[row[0]] = row[1]
                elif len(row) > 2 and row[2] == "yes":
                    sys.exit(f"{row[0]} is a required field, please complete the file {out_metadata} and re-run this step")

    print(f"\n\033[1m{step_counter.print_main_step()}: Preparing input for table2asn...\033[0m")
    list_ids = [args['genome']]
    in_fasta = os.path.join(args['input'], '03_CLUSTERING/', 'MVP_03_All_Sample_Filtered_Relaxed_Virus_Provirus_Sequences.fna')
    seq_genome = load_seq(in_fasta,list_ids)
    combined_description = f"[organism={tmp_metadata['organism']}]"
    combined_description = combined_description + f" [topology={tmp_metadata['topology']}]"
    combined_description = combined_description + f" [gcode={tmp_metadata['gcode']}]"
    combined_description = combined_description + f" [completedness={tmp_metadata['completedness']}]"
    combined_description = combined_description + f" [moltype={tmp_metadata['moltype']}]"
    combined_description = combined_description + f" [lat_lon=\"{tmp_metadata['lat_lon']}\"]"
    combined_description = combined_description + f" [collection_date={tmp_metadata['collection_date']}]"
    combined_description = combined_description + f" [metagenomic={tmp_metadata['metagenomic']}]"
    combined_description = combined_description + f" [environmental_sample={tmp_metadata['environmental_sample']}]"
    combined_description = combined_description + f" [isolation source=\"{tmp_metadata['geo_loc_name']}\"]"
    if "host" in tmp_metadata:
        combined_description = combined_description + f" [host={tmp_metadata['host']}]"
    if "sra_reads" in tmp_metadata:
        combined_description = combined_description + f" [note=genome assembled from reads in {tmp_metadata['sra_reads']}]"
    with open(out_fasta, "w") as output_handle:
        seq_record = SeqRecord(
            seq_genome[args['genome']],
            id=args['genome'],
            description=combined_description)
        SeqIO.write(seq_record, output_handle, "fasta")

    print(f"\n\033[1m{step_counter.print_main_step()}: Creating a clean version of the metadata...\033[0m")
    with open(out_metadata,"r") as f, open(clean_metadata,"w") as o:
        tag = 0
        tmp = {}
        freader = csv.reader(f, delimiter='\t', quoting=csv.QUOTE_NONE)
        mwriter = csv.writer(o, delimiter='\t',quotechar='"', quoting=csv.QUOTE_MINIMAL)
        for row in freader:
            if len(row) > 1 and row[1] != "":
                if row[0].startswith("StructuredCommentPrefix"):
                    tag = 1
                if tag == 1 and not(row[0].startswith("#")):
                    mwriter.writerow(row[0:2])
                    # print(f"{row[0]}\t{row[1]}")
                    if row[0] == "StructuredCommentPrefix" and row[1] == "Assembly-Data":
                        tab_asbly = tmp_metadata["assembly_software"].split(";")
                        reformat_asbly = tab_asbly[0] + " " + tab_asbly[1]
                        mwriter.writerow(["Assembly Method",reformat_asbly])

    # Finally copying over the annotation with the right file name so that it's picked up by table2asn
    shutil.copyfile(out_annotation, final_annotation)

    print(f"\n\033[1m{step_counter.print_main_step()}: Preparing to run table2asn as follows...\033[0m")
    cmd = f"table2asn -i {out_fasta} -w {clean_metadata} -t {template_file} -a s -V vb -Z "
    print(f"{cmd}")
    p = sp.Popen(cmd, shell=True)
    p.wait()
    # Testing if we have any error and reportin them here if we do
    print("##########################################################\n")
    print("List of warning / information reported by table2asn -- these are just for your information:")
    error_file = os.path.join(sub_dir, args['genome'] + "_genome.val")
    with open(error_file,"r") as f:
        for line in f.readlines():
            if line.startswith("Warning") or line.startswith("Info"):
                print(f"{line}")
    print("##########################################################\n")
    print("Checking for any errors reported by table2asn -- these would be important to fix before submission to NCBI:")
    error_file = os.path.join(sub_dir, args['genome'] + "_genome.val")
    tag = 0
    with open(error_file,"r") as f:
        for line in f.readlines():
            if not(line.startswith("Warning")) and not(line.startswith("Info")):
                print(f"UNEXPECTED ERROR -- {line}")
                tag = 1
    if tag == 0:
        print("MVP did not notice any major error here, that's great ! ")
    print("\n##########################################################\n\n")
    sqn_file = os.path.join(sub_dir, args['genome'] + "_genome.sqn")
    gb_file = os.path.join(sub_dir, args['genome'] + "_genome.gb")

    return sqn_file, gb_file



def main(args):
    # Capture start time
    start_time = datetime.datetime.now()
    formatted_start_time = start_time.strftime("%Y-%m-%d %H:%M:%S")
    print(f"\nStart Time: {formatted_start_time}\n")

    tpl_dir, sub_dir = create_output_directory(args)

    template_metadata = create_template_path()

    out_metadata, out_annotation, metadata, summary_file, seq_summary = load_files(args, tpl_dir)

    # If step = setup_metadata
    if args['step'] == "setup_metadata":
        set_up_metadata(args, seq_summary, summary_file, metadata, template_metadata, out_metadata, out_annotation, tpl_dir)

        # Capture end time
        end_time = datetime.datetime.now()
        formatted_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")

        # Calculate duration
        duration = end_time - start_time

        message1 = "\033[1mModule 99 finished: Submission files generated!\033[0m\n"
        line_of_stars = '*' * len(message1)
        print()
        print(line_of_stars)
        print(message1)
        print(f"Start Time: {formatted_start_time}") 
        print(f"End Time: {formatted_end_time}")
        print(f"Running Time: {duration.total_seconds():.2f} seconds\n")
        print(f"The files were created in {tpl_dir}, the next step is for you to review and complete the information in {out_metadata}")
        print(f"In particular, make sure that all rows noted as 'yes' in the 'required' column are filled in, and review (and modify if needed) all the rows pre-filled by MVP")
        print(f"Once this is done, you should run the step 'prep_submission' for the same genome, i.e. ")
        print(f"99_prep_MIUViG_submission.py -s prep_submission -t /path/to/sample/template.sbt -i {args['input']} -m {args['metadata']} -g {args['genome']}")
        print(line_of_stars)
        print()


    # If step = prep_submission
    if args['step'] == "prep_submission":
        sqn_file, gb_file = prep_submission(args, out_metadata, out_annotation, sub_dir)

        # Capture end time
        end_time = datetime.datetime.now()
        formatted_end_time = end_time.strftime("%Y-%m-%d %H:%M:%S")

        # Calculate duration
        duration = end_time - start_time

        message1 = "\033[1mModule 99 finished: Submission files generated!\033[0m\n"
        line_of_stars = '*' * len(message1)
        print()
        print(line_of_stars)
        print(message1)
        print(f"Start Time: {formatted_start_time}") 
        print(f"End Time: {formatted_end_time}")
        print(f"Running Time: {duration.total_seconds():.2f} seconds\n")
        print("Submission files generated, if you had some 'ERROR -- error' reported above, please correct the input files manually and re-run the prep_submission step or just table2asn using the command line listed above")
        print(f"To check the final annotated file, you can use {gb_file} in a genome browser")
        print(f"This GenBank file should include all the right information in the header, the MIUViG-formatted metadata as a formatted comment, and the gene annotation generated by geNomad")
        print(f"For final submission, you can use {sqn_file}")
        print(line_of_stars)
        print()
